import json, os, gzip
import unittest
import networkx as nw
import numpy as np
import datetime

from qiskit.providers.fake_provider import FakeVigo
from qiskit.providers.aer.noise import NoiseModel
from qiskit.providers.aer import QasmSimulator

from openqaoa import QAOA, RQAOA
from openqaoa.algorithms import QAOAResult, RQAOAResult
from openqaoa.algorithms.baseworkflow import Workflow
from openqaoa.utilities import X_mixer_hamiltonian, XY_mixer_hamiltonian, is_valid_uuid
from openqaoa.algorithms.workflow_properties import (
    BackendProperties,
    ClassicalOptimizer,
    CircuitProperties,
)
from openqaoa.algorithms.rqaoa.rqaoa_workflow_properties import RqaoaParameters
from openqaoa.backends import create_device, DeviceLocal
from openqaoa.backends.devices_core import SUPPORTED_LOCAL_SIMULATORS
from openqaoa.qaoa_components import (
    Hamiltonian,
    QAOADescriptor,
    QAOAVariationalStandardParams,
    QAOAVariationalStandardWithBiasParams,
    QAOAVariationalExtendedParams,
    QAOAVariationalFourierParams,
    QAOAVariationalFourierExtendedParams,
    QAOAVariationalFourierWithBiasParams,
)
from openqaoa.backends import QAOAvectorizedBackendSimulator
from openqaoa.backends.basebackend import QAOABaseBackendStatevector
from openqaoa.problems import MinimumVertexCover, QUBO, MaximumCut
from openqaoa.optimizers.qaoa_optimizer import available_optimizers
from openqaoa.optimizers.training_vqa import (
    ScipyOptimizer,
    CustomScipyGradientOptimizer,
    PennyLaneOptimizer,
)
from openqaoa_qiskit.backends import DeviceQiskit
from openqaoa_qiskit.backends import (
    QAOAQiskitBackendShotBasedSimulator,
    QAOAQiskitBackendStatevecSimulator,
)


ALLOWED_LOCAL_SIMUALTORS = SUPPORTED_LOCAL_SIMULATORS
LOCAL_DEVICES = ALLOWED_LOCAL_SIMUALTORS + ["6q-qvm", "Aspen-11"]


def _compare_qaoa_results(dict_old, dict_new):

    for key in dict_old.keys():
        if key == "cost_hamiltonian":  ## CHECK WHAT DO WITH THIS
            pass
        elif key == "_QAOAResult__type_backend":
            if issubclass(dict_old[key], QAOABaseBackendStatevector):
                assert (
                    dict_new[key] == QAOABaseBackendStatevector
                ), "Type of backend is not correct."
            else:
                assert dict_new[key] == "", "Type of backend should be empty string."
        elif key == "optimized":
            for key2 in dict_old[key].keys():
                if key2 == "measurement_outcomes":
                    assert np.all(
                        dict_old[key][key2] == dict_new[key][key2]
                    ), "Optimized params are not the same."
                else:
                    assert (
                        dict_old[key][key2] == dict_new[key][key2]
                    ), "Optimized params are not the same."
        elif key == "intermediate":
            for key2 in dict_old[key].keys():
                if key2 == "measurement_outcomes":
                    for step in range(len(dict_old[key][key2])):
                        assert np.all(
                            dict_old[key][key2][step] == dict_new[key][key2][step]
                        ), "Intermediate params are not the same."
                else:
                    assert (
                        dict_old[key][key2] == dict_new[key][key2]
                    ), "Intermediate params are not the same."
        else:
            assert dict_old[key] == dict_new[key], f"'{key}' is not the same"


def _test_keys_in_dict(obj, expected_keys):
    """
    private function to test the keys. It recursively tests the keys of the nested dictionaries, or lists of dictionaries
    """

    if isinstance(obj, dict):
        for key in obj:
            if key in expected_keys.keys():
                expected_keys[key] = True

            if isinstance(obj[key], dict):
                _test_keys_in_dict(obj[key], expected_keys)
            elif isinstance(obj[key], list):
                for item in obj[key]:
                    _test_keys_in_dict(item, expected_keys)
    elif isinstance(obj, list):
        for item in obj:
            _test_keys_in_dict(item, expected_keys)


class TestingVanillaQAOA(unittest.TestCase):

    """
    Unit test based testing of the QAOA workflow class
    """

    def test_set_device_cloud(self):
        """ "
        Check that all QPU-provider related devices are correctly initialised
        """
        q = QAOA()
        q.set_device(
            create_device(
                "ibmq",
                name="place_holder",
                **{"hub": "***", "group": "***", "project": "***"},
            )
        )
        assert type(q.device) == DeviceQiskit
        assert q.device.device_name == "place_holder"
        assert q.device.device_location == "ibmq"

    def test_compile_before_optimise(self):
        """
        Assert that compilation has to be called before optimisation
        """
        g = nw.circulant_graph(6, [1])
        # vc = MinimumVertexCover(g, field =1.0, penalty=10).qubo

        q = QAOA()
        q.set_classical_optimizer(optimization_progress=True)

        self.assertRaises(ValueError, lambda: q.optimize())

    def test_set_backend_properties_check_backend_qiskit_qasm(self):

        """
        Check if the backend returned by set_backend_properties is correct
        Based on the input device. For qiskit qasm simulator.
        Also Checks if defaults from workflows are used in the backend.
        """

        nodes = 6
        edge_probability = 0.6
        g = nw.generators.fast_gnp_random_graph(n=nodes, p=edge_probability)
        problem = MinimumVertexCover(g, field=1.0, penalty=10)
        qubo_problem = problem.qubo

        q = QAOA()
        q.set_device(create_device(location="local", name="qiskit.qasm_simulator"))
        q.compile(problem=qubo_problem)

        self.assertEqual(type(q.backend), QAOAQiskitBackendShotBasedSimulator)

        self.assertEqual(q.backend.init_hadamard, True)
        self.assertEqual(q.backend.prepend_state, None)
        self.assertEqual(q.backend.append_state, None)
        self.assertEqual(q.backend.cvar_alpha, 1)
        self.assertEqual(q.backend.n_shots, 100)

    def test_set_backend_properties_check_backend_qiskit_statevector(self):

        """
        Check if the backend returned by set_backend_properties is correct
        Based on the input device. For qiskit statevector simulator.
        Also Checks if defaults from workflows are used in the backend.
        """

        nodes = 6
        edge_probability = 0.6
        g = nw.generators.fast_gnp_random_graph(n=nodes, p=edge_probability)
        problem = MinimumVertexCover(g, field=1.0, penalty=10)
        qubo_problem = problem.qubo

        q = QAOA()
        q.set_device(
            create_device(location="local", name="qiskit.statevector_simulator")
        )
        q.compile(problem=qubo_problem)

        self.assertEqual(type(q.backend), QAOAQiskitBackendStatevecSimulator)

        self.assertEqual(q.backend.init_hadamard, True)
        self.assertEqual(q.backend.prepend_state, None)
        self.assertEqual(q.backend.append_state, None)
        self.assertEqual(q.backend.cvar_alpha, 1)

        self.assertRaises(AttributeError, lambda: q.backend.n_shots)

    def test_qaoa_asdict_with_noise(self):
        """test to check that we can serialize a QAOA object with noise"""
        device_backend = FakeVigo()
        device = QasmSimulator.from_backend(device_backend)
        noise_model = NoiseModel.from_backend(device)
        q_noisy_shot = QAOA()

        # device
        qiskit_noisy_shot = create_device(
            location="local", name="qiskit.qasm_simulator"
        )
        q_noisy_shot.set_device(qiskit_noisy_shot)
        # circuit properties
        q_noisy_shot.set_circuit_properties(
            p=2, param_type="standard", init_type="rand", mixer_hamiltonian="x"
        )
        # backend properties
        q_noisy_shot.set_backend_properties(n_shots=200, noise_model=noise_model)
        # classical optimizer properties
        q_noisy_shot.set_classical_optimizer(
            method="COBYLA", maxiter=200, cost_progress=True, parameter_log=True
        )
        q_noisy_shot.compile(QUBO.random_instance(n=8))
        q_noisy_shot.optimize()
        q_noisy_shot.asdict()

    def test_qaoa_from_dict_and_load(self):
        """
        test loading the QAOA object from a dictionary
        methods: from_dict, load, loads
        """

        # problem
        maxcut_qubo = MaximumCut(
            nw.generators.fast_gnp_random_graph(n=6, p=0.6, seed=42)
        ).qubo

        # run rqaoa with different devices, and save the objcets in a list
        qaoas = []
        for device in [
            create_device(location="local", name="qiskit.shot_simulator")
        ]:

            q = QAOA()
            q.set_device(device)
            q.set_circuit_properties(
                p=1, param_type="extended", init_type="rand", mixer_hamiltonian="x"
            )
            q.set_backend_properties(n_shots=50)
            q.set_classical_optimizer(maxiter=10, optimization_progress=True)
            q.set_exp_tags({"add_tag": "test"})
            q.set_header(
                project_id="8353185c-b175-4eda-9628-b4e58cb0e41b",
                description="test",
                run_by="raul",
                provider="-",
                target="-",
                cloud="local",
                client="-",
            )

            # test that you can convert the rqaoa object to a dictionary and then load it before optimization
            _ = QAOA.from_dict(q.asdict())
            _.compile(maxcut_qubo)
            _.optimize()
            assert isinstance(
                _, QAOA
            ), "The object loaded from a dictionary is not an RQAOA object."

            # compile and optimize the original rqaoa object
            q.compile(maxcut_qubo)
            q.optimize()

            qaoas.append(q)

        # for each rqaoa object, create a new rqaoa object from dict, json string, json file, and compressed json file and compare them with the original object
        for q in qaoas:

            new_q_list = []

            # get new qaoa from dict
            new_q_list.append(QAOA.from_dict(q.asdict()))
            # get new qaoa from json string
            new_q_list.append(QAOA.loads(q.dumps()))
            # get new qaoa from json file
            q.dump("test.json", prepend_id=False)
            new_q_list.append(QAOA.load("test.json"))
            os.remove("test.json")  # delete file test.json
            # get new qaoa from compressed json file
            q.dump("test.json", prepend_id=False, compresslevel=3)
            new_q_list.append(QAOA.load("test.json.gz"))
            os.remove("test.json.gz")  # delete file test.json

            for new_q in new_q_list:

                # check that the new object is an QAOA object
                assert isinstance(new_q, QAOA), "new_r is not an RQAOA object"

                # check that the attributes of the new object are of the correct type
                attributes_types = [
                    ("header", dict),
                    ("exp_tags", dict),
                    ("problem", QUBO),
                    ("result", QAOAResult),
                    ("backend_properties", BackendProperties),
                    ("classical_optimizer", ClassicalOptimizer),
                    ("circuit_properties", CircuitProperties),
                ]
                for attribute, type_ in attributes_types:
                    assert isinstance(
                        getattr(new_q, attribute), type_
                    ), f"attribute {attribute} is not type {type_}"

                # get the two objects (old and new) as dictionaries
                q_asdict = q.asdict()
                new_q_asdict = new_q.asdict()

                # compare the two dictionaries
                for key, value in q_asdict.items():
                    if key == "header":
                        assert value == new_q_asdict[key], "Header is not the same"

                    elif key == "data":
                        for key2, value2 in value.items():
                            if key2 == "input_parameters":
                                # pop key device since it is not returned completely when using asdict/dump(s)
                                value2.pop("device")
                                new_q_asdict[key][key2].pop("device")
                            if key2 == "result":
                                _compare_qaoa_results(value2, new_q_asdict[key][key2])
                            else:
                                assert (
                                    value2 == new_q_asdict[key][key2]
                                ), "{} not the same".format(key2)

                # compile and optimize the new qaoa, to check if everything is working
                new_q.compile(maxcut_qubo)
                new_q.optimize()

        # check that the RQAOA.from_dict method raises an error when using a QAOA dictionary
        error = False
        try:
            RQAOA.from_dict(q.asdict())
        except Exception:
            error = True
        assert (
            error
        ), "RQAOA.from_dict should raise an error when using a QAOA dictionary"


class TestingRQAOA(unittest.TestCase):
    """
    Unit test based testing of the RQAOA workflow class
    """

    def _test_default_values(self, x):
        """
        General function to check default values of rqaoa and qaoa
        """

        # circuit_properties
        cp = x.circuit_properties
        assert cp.param_type == "standard"
        assert cp.init_type == "ramp"
        assert cp.p == 1
        assert cp.q == None
        assert cp.mixer_hamiltonian == "x"

        # device
        d = x.device
        assert d.device_location == "local"
        assert d.device_name == "vectorized"

    def test_rqaoa_default_values(self):
        """
        Tests all default values are correct
        """
        r = RQAOA()

        assert r.rqaoa_parameters.rqaoa_type == "custom"
        assert r.rqaoa_parameters.n_cutoff == 5
        assert r.rqaoa_parameters.n_max == 1
        assert r.rqaoa_parameters.steps == 1
        assert r.rqaoa_parameters.original_hamiltonian == None
        assert r.rqaoa_parameters.counter == 0

        self._test_default_values(r)

    def test_rqaoa_compile_and_qoao_default_values(self):
        """
        Test creation of the qaoa object and its default values
        """
        r = RQAOA()
        r.compile(QUBO.random_instance(n=7))

        self._test_default_values(r._RQAOA__q)

    def __run_rqaoa(
        self,
        type,
        problem=None,
        n_cutoff=5,
        eliminations=1,
        p=1,
        param_type="standard",
        mixer="x",
        method="cobyla",
        maxiter=15,
        name_device="qiskit.statevector_simulator",
        return_object=False,
    ):

        if problem == None:
            problem = MaximumCut.random_instance(
                n_nodes=8, edge_probability=0.5, seed=2
            ).qubo

        r = RQAOA()
        qiskit_device = create_device(location="local", name=name_device)
        r.set_device(qiskit_device)
        if type == "adaptive":
            r.set_rqaoa_parameters(
                n_cutoff=n_cutoff, n_max=eliminations, rqaoa_type=type
            )
        else:
            r.set_rqaoa_parameters(
                n_cutoff=n_cutoff, steps=eliminations, rqaoa_type=type
            )
        r.set_circuit_properties(p=p, param_type=param_type, mixer_hamiltonian=mixer)
        r.set_backend_properties(prepend_state=None, append_state=None)
        r.set_classical_optimizer(
            method=method,
            maxiter=maxiter,
            optimization_progress=True,
            cost_progress=True,
            parameter_log=True,
        )
        r.set_header(
            project_id="8353185c-b175-4eda-9628-b4e58cb0e41b",
            description="header",
            run_by="raul",
            provider="-",
            target="-",
            cloud="local",
            client="-",
        )
        r.set_exp_tags(tags={"tag1": "value1", "tag2": "value2"})
        r.compile(problem)
        r.optimize()

        if return_object:
            return r
        return r.result.get_solution()

    def test_rqaoa_optimize_multiple_times(self):
        """
        Test that the rqaoa can not be optimized multiple times
        """
        graph = nw.circulant_graph(10, [1])
        problem = MinimumVertexCover(graph, field=1, penalty=10).qubo

        r = RQAOA()
        exception = False
        try:
            r.optimize()
        except:
            exception = True

        assert exception, "RQAOA should not be able to optimize without compilation"

        r.compile(problem)
        r.optimize()
        exception = False
        try:
            r.optimize()
        except:
            exception = True

        assert (
            exception
        ), "RQAOA should not be able to optimize twice without compilation"

    def test_example_1_adaptive_custom(self):

        # Number of qubits
        n_qubits = 12

        # Elimination schemes
        Nmax = [1, 2, 3, 4]
        schedules = [1, [1, 2, 1, 2, 7]]
        n_cutoff = 5

        # Edges and weights of the graph
        pair_edges = [(i, i + 1) for i in range(n_qubits - 1)] + [(0, n_qubits - 1)]
        self_edges = [(i,) for i in range(n_qubits)]
        pair_weights = [1 for _ in range(len(pair_edges))]  # All weights equal to 1
        self_weights = [10 ** (-4) for _ in range(len(self_edges))]

        edges = pair_edges + self_edges
        weights = pair_weights + self_weights
        problem = QUBO(n_qubits, edges, weights)

        # list of solutions of rqaoa
        solutions = []

        # run RQAOA and append solution
        for nmax in Nmax:
            solutions.append(self.__run_rqaoa("adaptive", problem, n_cutoff, nmax))

        for schedule in schedules:
            solutions.append(self.__run_rqaoa("custom", problem, n_cutoff, schedule))

        # Correct solution
        exact_soutions = {"101010101010": -12, "010101010101": -12}

        # Check computed solutions are among the correct ones
        for solution in solutions:
            for key in solution:
                assert solution[key] == exact_soutions[key]

    def test_example_2_adaptive_custom(self):

        # Elimination scheme
        n_cutoff = 3

        # Define problem instance (Ring graph 10 qubits)
        graph = nw.circulant_graph(10, [1])
        problem = MinimumVertexCover(graph, field=1, penalty=10).qubo

        # run RQAOA and append solution in list
        solutions = []
        solutions.append(self.__run_rqaoa("adaptive", problem, n_cutoff))
        solutions.append(self.__run_rqaoa("custom", problem, n_cutoff))

        # Correct solution
        exact_soutions = {"1010101010": 5, "0101010101": 5}

        # Check computed solutions are among the correct ones
        for solution in solutions:
            for key in solution:
                assert solution[key] == exact_soutions[key]

    def test_example_3_adaptive_custom(self):

        # Elimination scheme
        step = 2
        nmax = 4
        n_cutoff = 3

        # Define problem instance (Ring graph 10 qubits)
        graph = nw.complete_graph(10)
        problem = MinimumVertexCover(graph, field=1, penalty=10).qubo

        # run RQAOA and append solution in list
        solutions = []
        solutions.append(self.__run_rqaoa("adaptive", problem, n_cutoff, nmax))
        solutions.append(self.__run_rqaoa("custom", problem, n_cutoff, step))

        # Correct solution
        exact_soutions = {
            "0111111111": 9,
            "1011111111": 9,
            "1101111111": 9,
            "1110111111": 9,
            "1111011111": 9,
            "1111101111": 9,
            "1111110111": 9,
            "1111111011": 9,
            "1111111101": 9,
            "1111111110": 9,
        }

        # Check computed solutions are among the correct ones
        for solution in solutions:
            for key in solution:
                assert solution[key] == exact_soutions[key]

    def test_example_4_adaptive_custom(self):

        # Number of qubits
        n_qubits = 10

        # Elimination schemes
        Nmax = [1, 2, 3, 4]
        schedules = [1, 2, 3]
        n_cutoff = 3

        # Edges and weights of the graph
        edges = [(i, j) for j in range(n_qubits) for i in range(j)]
        weights = [1 for _ in range(len(edges))]

        problem = QUBO(n_qubits, edges, weights)

        # list of solutions of rqaoa
        solutions = []

        # run RQAOA and append solution
        for nmax in Nmax:
            solutions.append(self.__run_rqaoa("adaptive", problem, n_cutoff, nmax))

        for schedule in schedules:
            solutions.append(self.__run_rqaoa("custom", problem, n_cutoff, schedule))

        # Correct solution
        exact_states = [
            "1111100000",
            "1111010000",
            "1110110000",
            "1101110000",
            "1011110000",
            "0111110000",
            "1111001000",
            "1110101000",
            "1101101000",
            "1011101000",
            "0111101000",
            "1110011000",
            "1101011000",
            "1011011000",
            "0111011000",
            "1100111000",
            "1010111000",
            "0110111000",
            "1001111000",
            "0101111000",
            "0011111000",
            "1111000100",
            "1110100100",
            "1101100100",
            "1011100100",
            "0111100100",
            "1110010100",
            "1101010100",
            "1011010100",
            "0111010100",
            "1100110100",
            "1010110100",
            "0110110100",
            "1001110100",
            "0101110100",
            "0011110100",
            "1110001100",
            "1101001100",
            "1011001100",
            "0111001100",
            "1100101100",
            "1010101100",
            "0110101100",
            "1001101100",
            "0101101100",
            "0011101100",
            "1100011100",
            "1010011100",
            "0110011100",
            "1001011100",
            "0101011100",
            "0011011100",
            "1000111100",
            "0100111100",
            "0010111100",
            "0001111100",
            "1111000010",
            "1110100010",
            "1101100010",
            "1011100010",
            "0111100010",
            "1110010010",
            "1101010010",
            "1011010010",
            "0111010010",
            "1100110010",
            "1010110010",
            "0110110010",
            "1001110010",
            "0101110010",
            "0011110010",
            "1110001010",
            "1101001010",
            "1011001010",
            "0111001010",
            "1100101010",
            "1010101010",
            "0110101010",
            "1001101010",
            "0101101010",
            "0011101010",
            "1100011010",
            "1010011010",
            "0110011010",
            "1001011010",
            "0101011010",
            "0011011010",
            "1000111010",
            "0100111010",
            "0010111010",
            "0001111010",
            "1110000110",
            "1101000110",
            "1011000110",
            "0111000110",
            "1100100110",
            "1010100110",
            "0110100110",
            "1001100110",
            "0101100110",
            "0011100110",
            "1100010110",
            "1010010110",
            "0110010110",
            "1001010110",
            "0101010110",
            "0011010110",
            "1000110110",
            "0100110110",
            "0010110110",
            "0001110110",
            "1100001110",
            "1010001110",
            "0110001110",
            "1001001110",
            "0101001110",
            "0011001110",
            "1000101110",
            "0100101110",
            "0010101110",
            "0001101110",
            "1000011110",
            "0100011110",
            "0010011110",
            "0001011110",
            "0000111110",
            "1111000001",
            "1110100001",
            "1101100001",
            "1011100001",
            "0111100001",
            "1110010001",
            "1101010001",
            "1011010001",
            "0111010001",
            "1100110001",
            "1010110001",
            "0110110001",
            "1001110001",
            "0101110001",
            "0011110001",
            "1110001001",
            "1101001001",
            "1011001001",
            "0111001001",
            "1100101001",
            "1010101001",
            "0110101001",
            "1001101001",
            "0101101001",
            "0011101001",
            "1100011001",
            "1010011001",
            "0110011001",
            "1001011001",
            "0101011001",
            "0011011001",
            "1000111001",
            "0100111001",
            "0010111001",
            "0001111001",
            "1110000101",
            "1101000101",
            "1011000101",
            "0111000101",
            "1100100101",
            "1010100101",
            "0110100101",
            "1001100101",
            "0101100101",
            "0011100101",
            "1100010101",
            "1010010101",
            "0110010101",
            "1001010101",
            "0101010101",
            "0011010101",
            "1000110101",
            "0100110101",
            "0010110101",
            "0001110101",
            "1100001101",
            "1010001101",
            "0110001101",
            "1001001101",
            "0101001101",
            "0011001101",
            "1000101101",
            "0100101101",
            "0010101101",
            "0001101101",
            "1000011101",
            "0100011101",
            "0010011101",
            "0001011101",
            "0000111101",
            "1110000011",
            "1101000011",
            "1011000011",
            "0111000011",
            "1100100011",
            "1010100011",
            "0110100011",
            "1001100011",
            "0101100011",
            "0011100011",
            "1100010011",
            "1010010011",
            "0110010011",
            "1001010011",
            "0101010011",
            "0011010011",
            "1000110011",
            "0100110011",
            "0010110011",
            "0001110011",
            "1100001011",
            "1010001011",
            "0110001011",
            "1001001011",
            "0101001011",
            "0011001011",
            "1000101011",
            "0100101011",
            "0010101011",
            "0001101011",
            "1000011011",
            "0100011011",
            "0010011011",
            "0001011011",
            "0000111011",
            "1100000111",
            "1010000111",
            "0110000111",
            "1001000111",
            "0101000111",
            "0011000111",
            "1000100111",
            "0100100111",
            "0010100111",
            "0001100111",
            "1000010111",
            "0100010111",
            "0010010111",
            "0001010111",
            "0000110111",
            "1000001111",
            "0100001111",
            "0010001111",
            "0001001111",
            "0000101111",
            "0000011111",
        ]
        exact_soutions = {state: -5 for state in exact_states}

        # Check computed solutions are among the correct ones
        for solution in solutions:
            for key in solution:
                assert solution[key] == exact_soutions[key]

    def test_rqaoa_asdict_dumps(self):
        """Test the asdict method of the RQAOA class."""

        # rqaoa
        rqaoa = self.__run_rqaoa("custom", return_object=True)

        # check RQAOA asdict
        self.__test_expected_keys(rqaoa.asdict(), method="asdict")

        # check RQAOA asdict deleting some keys
        exclude_keys = ["corr_matrix", "number_steps"]
        self.__test_expected_keys(
            rqaoa.asdict(exclude_keys=exclude_keys), exclude_keys, method="asdict"
        )

        # check RQAOA dumps
        self.__test_expected_keys(json.loads(rqaoa.dumps()), method="dumps")

        # check RQAOA dumps deleting some keys
        exclude_keys = ["project_id", "counter"]
        self.__test_expected_keys(
            json.loads(rqaoa.dumps(exclude_keys=exclude_keys)),
            exclude_keys,
            method="dumps",
        )

        # check RQAOA dump
        file_name = "test_dump_rqaoa.json"
        project_id, experiment_id, atomic_id = (
            rqaoa.header["project_id"],
            rqaoa.header["experiment_id"],
            rqaoa.header["atomic_id"],
        )
        full_name = f"{project_id}--{experiment_id}--{atomic_id}--{file_name}"

        rqaoa.dump(file_name, prepend_id=True, indent=None)
        assert os.path.isfile(full_name), "Dump file does not exist"
        with open(full_name, "r") as file:
            assert file.read() == rqaoa.dumps(
                indent=None
            ), "Dump file does not contain the correct data"
        os.remove(full_name)

        # check RQAOA dump whitout prepending the experiment_id and atomic_id
        rqaoa.dump(file_name, indent=None, prepend_id=False)
        assert os.path.isfile(
            file_name
        ), "Dump file does not exist, when not prepending the experiment_id and atomic_id"

        # check RQAOA dump fails when the file already exists
        error = False
        try:
            rqaoa.dump(file_name, indent=None, prepend_id=False)
        except FileExistsError:
            error = True
        assert error, "Dump file does not fail when the file already exists"

        # check that we can overwrite the file
        rqaoa.dump(file_name, indent=None, prepend_id=False, overwrite=True)
        assert os.path.isfile(file_name), "Dump file does not exist, when overwriting"
        os.remove(file_name)

        # check RQAOA dump fails when prepend_id is True and file_name is not given
        error = False
        try:
            rqaoa.dump(prepend_id=False)
        except ValueError:
            error = True
        assert (
            error
        ), "Dump file does not fail when prepend_id is True and file_name is not given"

        # check RQAOA dump with no arguments
        error = False
        try:
            rqaoa.dump()
        except ValueError:
            error = True
        assert (
            error
        ), "Dump file does not fail when no arguments are given, should be the same as dump(prepend_id=False)"

        # check you can dump to a file with no arguments, just prepend_id=True
        rqaoa.dump(prepend_id=True)
        assert os.path.isfile(
            f"{project_id}--{experiment_id}--{atomic_id}.json"
        ), "Dump file does not exist, when no name is given"
        os.remove(f"{project_id}--{experiment_id}--{atomic_id}.json")

        # check RQAOA dump deleting some keys
        exclude_keys = ["schedule", "singlet"]
        rqaoa.dump(file_name, exclude_keys=exclude_keys, indent=None, prepend_id=True)
        assert os.path.isfile(
            full_name
        ), "Dump file does not exist, when deleting some keys"
        with open(full_name, "r") as file:
            assert file.read() == rqaoa.dumps(
                exclude_keys=exclude_keys, indent=None
            ), "Dump file does not contain the correct data, when deleting some keys"
        os.remove(full_name)

        # check RQAOA dump with compression
        rqaoa.dump(file_name, compresslevel=2, indent=None, prepend_id=True)
        assert os.path.isfile(
            full_name + ".gz"
        ), "Dump file does not exist, when compressing"
        with gzip.open(full_name + ".gz", "rb") as file:
            assert (
                file.read() == rqaoa.dumps(indent=None).encode()
            ), "Dump file does not contain the correct data, when compressing"
        os.remove(full_name + ".gz")


    def __test_expected_keys(self, obj, exclude_keys=[], method="asdict"):
        """
        method to test if the dictionary has all the expected keys
        """

        # create a dictionary with all the expected keys and set them to False
        expected_keys = [
            "header",
            "atomic_id",
            "experiment_id",
            "project_id",
            "algorithm",
            "description",
            "run_by",
            "provider",
            "target",
            "cloud",
            "client",
            "qubit_number",
            "execution_time_start",
            "execution_time_end",
            "metadata",
            "tag1",
            "tag2",
            "problem_type",
            "n_shots",
            "optimizer_method",
            "param_type",
            "init_type",
            "p",
            "rqaoa_type",
            "rqaoa_n_max",
            "rqaoa_n_cutoff",
            "data",
            "exp_tags",
            "input_problem",
            "terms",
            "weights",
            "constant",
            "n",
            "problem_instance",
            "input_parameters",
            "device",
            "device_location",
            "device_name",
            "backend_properties",
            "init_hadamard",
            "prepend_state",
            "append_state",
            "cvar_alpha",
            "noise_model",
            "initial_qubit_mapping",
            "seed_simulator",
            "qiskit_simulation_method",
            "active_reset",
            "rewiring",
            "disable_qubit_rewiring",
            "classical_optimizer",
            "optimize",
            "method",
            "maxiter",
            "maxfev",
            "jac",
            "hess",
            "constraints",
            "bounds",
            "tol",
            "optimizer_options",
            "jac_options",
            "hess_options",
            "parameter_log",
            "optimization_progress",
            "cost_progress",
            "save_intermediate",
            "circuit_properties",
            "qubit_register",
            "q",
            "variational_params_dict",
            "total_annealing_time",
            "annealing_time",
            "linear_ramp_time",
            "mixer_hamiltonian",
            "mixer_qubit_connectivity",
            "mixer_coeffs",
            "seed",
            "rqaoa_parameters",
            "n_max",
            "steps",
            "n_cutoff",
            "original_hamiltonian",
            "counter",
            "result",
            "solution",
            "classical_output",
            "minimum_energy",
            "optimal_states",
            "elimination_rules",
            "pair",
            "correlation",
            "schedule",
            "number_steps",
            "intermediate_steps",
            "problem",
            "qaoa_results",
            "evals",
            "number_of_evals",
            "jac_evals",
            "qfim_evals",
            "most_probable_states",
            "solutions_bitstrings",
            "bitstring_energy",
            "intermediate",
            "angles",
            "cost",
            "measurement_outcomes",
            "job_id",
            "optimized",
            "eval_number",
            "exp_vals_z",
            "corr_matrix",
            "atomic_ids",
        ]
        expected_keys = {item: False for item in expected_keys}

        # test the keys, it will set the keys to True if they are found
        _test_keys_in_dict(obj, expected_keys)

        # Check if the dictionary has all the expected keys except the ones that were not included
        for key, value in expected_keys.items():
            if key not in exclude_keys:
                assert (
                    value == True
                ), f'Key "{key}" not found in the dictionary, when using "{method}" method.'
            else:
                assert (
                    value == False
                ), f'Key "{key}" was found in the dictionary, but it should not be there, when using "{method}" method.'

        """
        to get the list of expected keys, run the following code:

            def get_keys(obj, list_keys):
                if isinstance(obj, dict):
                    for key in obj:
                        if not key in list_keys: list_keys.append(key)

                        if isinstance(obj[key], dict):
                            get_keys(obj[key], list_keys)
                        elif isinstance(obj[key], list):
                            for item in obj[key]:
                                get_keys(item, list_keys)
                elif isinstance(obj, list):
                    for item in obj:
                        get_keys(item, list_keys)

            expected_keys = []
            get_keys(rqaoa.asdict(), expected_keys)
            print(expected_keys)
        """

    def test_rqaoa_dumping_step_by_step(self):
        """
        test dumping the RQAOA object step by step
        """

        # define the problem
        problem = QUBO.random_instance(n=8)
        problem.set_metadata(
            {"metadata_key1": "metadata_value1", "metadata_key2": "metadata_value2"}
        )

        # define the RQAOA object
        r = RQAOA()

        # set experimental tags
        r.set_exp_tags({"tag1": "value1", "tag2": "value2"})

        # set the classical optimizer
        r.set_classical_optimizer(optimization_progress=True)

        # compile the problem
        r.compile(problem)

        # optimize the problem while dumping the data at each step
        r.optimize(
            dump=True,
            dump_options={
                "file_name": "test_dumping_step_by_step",
                "compresslevel": 2,
                "indent": None,
                "prepend_id": True,
            },
        )

        # create list of expected file names
        project_id = "None"
        experiment_id, atomic_id = r.header["experiment_id"], r.header["atomic_id"]
        file_names = {
            id: project_id
            + "--"
            + experiment_id
            + "--"
            + id
            + "--"
            + "test_dumping_step_by_step.json.gz"
            for id in r.result["atomic_ids"].values()
        }
        file_names[atomic_id] = (
            project_id
            + "--"
            + experiment_id
            + "--"
            + atomic_id
            + "--"
            + "test_dumping_step_by_step.json.gz"
        )

        # check if the files exist
        for file_name in file_names.values():
            assert os.path.isfile(file_name), f"File {file_name} does not exist."

        # put each file in a dictionary
        files = {}
        for atomic_id, file_name in file_names.items():
            with gzip.open(file_name, "rb") as file:
                files[atomic_id] = json.loads(file.read().decode())

        rqaoa_files, qaoa_files = 0, 0

        # check if the files have the expected keys
        for atomic_id, dictionary in files.items():

            file_name = file_names[atomic_id]

            if r.header["atomic_id"] == atomic_id:  # rqaoa files

                rqaoa_files += 1

                assert (
                    dictionary["header"]["experiment_id"] == r.header["experiment_id"]
                ), f"File {file_name} has a different experiment_id than the RQAOA object."
                assert (
                    dictionary["header"]["atomic_id"] == r.header["atomic_id"]
                ), f"File {file_name} has a different atomic_id than the RQAOA object."
                assert (
                    dictionary["header"]["algorithm"] == "rqaoa"
                ), f"File {file_name} has a different algorithm than rqaoa, which is the expected algorithm."

                # check that the intermediate mesuraments are empty
                for step in dictionary["data"]["result"]["intermediate_steps"]:
                    assert (
                        step["qaoa_results"]["intermediate"]["measurement_outcomes"]
                        == []
                    ), f"File {file_name} has intermediate mesuraments, but it should not have them."

            else:  # qaoa files

                qaoa_files += 1

                assert (
                    dictionary["header"]["atomic_id"] == atomic_id
                ), f"File {file_name} has a different atomic_id than expected."
                assert (
                    dictionary["header"]["algorithm"] == "qaoa"
                ), f"File {file_name} has a different algorithm than qaoa, which is the expected algorithm."

                # check that the intermediate mesuraments are not empty
                assert (
                    len(
                        dictionary["data"]["result"]["intermediate"][
                            "measurement_outcomes"
                        ]
                    )
                    > 0
                ), f"File {file_name} does not have intermediate mesuraments, but it should have them."

        assert rqaoa_files == 1, f"Expected 1 rqaoa file, but {rqaoa_files} were found."
        assert qaoa_files == len(
            r.result["atomic_ids"]
        ), f'Expected {len(r.result["atomic_ids"])} qaoa files, but {qaoa_files} were found.'

        # erease the files
        for file_name in file_names.values():
            os.remove(file_name)

    def test_rqaoa_from_dict_and_load(self):
        """
        test loading the RQAOA object from a dictionary
        methods: from_dict, load, loads
        """

        # problem
        maxcut_qubo = MaximumCut(
            nw.generators.fast_gnp_random_graph(n=6, p=0.6, seed=42)
        ).qubo

        # run rqaoa with different devices, and save the objcets in a list
        rqaoas = []
        for device in [
            create_device(location="local", name="qiskit.shot_simulator"),
        ]:

            r = RQAOA()
            r.set_device(device)
            r.set_circuit_properties(
                p=1, param_type="extended", init_type="rand", mixer_hamiltonian="x"
            )
            r.set_backend_properties(n_shots=50)
            r.set_classical_optimizer(maxiter=10, optimization_progress=True)
            r.set_rqaoa_parameters(rqaoa_type="adaptive", n_cutoff=3)
            r.set_exp_tags({"tag1": "value1", "tag2": "value2"})
            r.set_header(
                project_id="8353185c-b175-4eda-9628-b4e58cb0e41b",
                description="test",
                run_by="raul",
                provider="-",
                target="-",
                cloud="local",
                client="-",
            )

            # test that you can convert the rqaoa object to a dictionary and then load it before optimization
            _ = RQAOA.from_dict(r.asdict())
            _.compile(maxcut_qubo)
            _.optimize()
            assert isinstance(
                _, RQAOA
            ), "The object loaded from a dictionary is not an RQAOA object."

            # compile and optimize the original rqaoa object
            r.compile(maxcut_qubo)
            r.optimize()

            rqaoas.append(r)

        # for each rqaoa object, create a new rqaoa object from dict, json string, json file, and compressed json file and compare them with the original object
        for r in rqaoas:

            new_r_list = []

            # get new qaoa from dict
            new_r_list.append(RQAOA.from_dict(r.asdict()))
            # get new qaoa from json string
            new_r_list.append(RQAOA.loads(r.dumps()))
            # get new qaoa from json file
            r.dump("test.json", prepend_id=False)
            new_r_list.append(RQAOA.load("test.json"))
            os.remove("test.json")  # delete file test.json
            # get new qaoa from compressed json file
            r.dump("test.json", prepend_id=False, compresslevel=3)
            new_r_list.append(RQAOA.load("test.json.gz"))
            os.remove("test.json.gz")  # delete file test.json

            for new_r in new_r_list:

                # check that the new object is an RQAOA object
                assert isinstance(new_r, RQAOA), "new_r is not an RQAOA object"

                # check that the attributes of the new object are of the correct type
                attributes_types = [
                    ("header", dict),
                    ("exp_tags", dict),
                    ("problem", QUBO),
                    ("result", RQAOAResult),
                    ("backend_properties", BackendProperties),
                    ("classical_optimizer", ClassicalOptimizer),
                    ("circuit_properties", CircuitProperties),
                    ("rqaoa_parameters", RqaoaParameters),
                ]
                for attribute, type_ in attributes_types:
                    assert isinstance(
                        getattr(new_r, attribute), type_
                    ), f"attribute {attribute} is not type {type_}"

                # get the two objects (old and new) as dictionaries
                r_asdict = r.asdict()
                new_r_asdict = new_r.asdict()

                # compare the two dictionaries
                for key, value in r_asdict.items():
                    if key == "header":
                        assert value == new_r_asdict[key], "Header is not the same"

                    elif key == "data":
                        for key2, value2 in value.items():
                            if key2 == "input_parameters":
                                # pop key device
                                value2.pop("device")
                                new_r_asdict[key][key2].pop("device")
                            if key2 == "result":
                                for step in range(len(value2["intermediate_steps"])):
                                    for key3 in value2["intermediate_steps"][
                                        step
                                    ].keys():
                                        if key3 == "qaoa_results":
                                            _compare_qaoa_results(
                                                value2["intermediate_steps"][step][
                                                    "qaoa_results"
                                                ],
                                                new_r_asdict[key][key2][
                                                    "intermediate_steps"
                                                ][step]["qaoa_results"],
                                            )
                                        else:
                                            assert (
                                                value2["intermediate_steps"][step][key3]
                                                == new_r_asdict[key][key2][
                                                    "intermediate_steps"
                                                ][step][key3]
                                            ), f"{key3} is not the same"
                            else:
                                assert (
                                    value2 == new_r_asdict[key][key2]
                                ), "{} not the same".format(key2)

                # compile and optimize the new rqaoa, to check if everything is working
                new_r.compile(maxcut_qubo)
                new_r.optimize()

        # check that the Optimizer.from_dict method raises an error when using a RQAOA dictionary
        error = False
        try:
            Workflow.from_dict(r.asdict())
        except Exception:
            error = True
        assert (
            error
        ), "Optimizer.from_dict should raise an error when using a RQAOA dictionary"


if __name__ == "__main__":
    unittest.main()



import unittest
import networkx as nx
import numpy as np
import matplotlib.pyplot as plt
from docplex.mp.model import Model

from openqaoa import QAOA
from openqaoa.problems import MinimumVertexCover
from openqaoa.qaoa_components import PauliOp, Hamiltonian
from openqaoa.algorithms.qaoa.qaoa_result import QAOAResult, most_probable_bitstring
from openqaoa.utilities import qaoa_probabilities
from openqaoa.problems.converters import FromDocplex2IsingModel
from openqaoa.backends import create_device


class TestingLoggerClass(unittest.TestCase):
    def test_attribute_existence(self):
        """
        This test checks that 'most_probable_states', 'evals', 'intermediate', 'optimized' are correctly
        created by the default workflow
        """

        # Create the problem
        g = nx.circulant_graph(6, [1])
        vc = MinimumVertexCover(g, field=1.0, penalty=10).qubo

        q = QAOA()
        q.compile(vc, verbose=False)
        q.optimize()

        assert set(
            ["most_probable_states", "evals", "intermediate", "optimized"]
        ) <= set([a for a in dir(q.result) if not a.startswith("__")])

    def test_most_probable_bitstring(self):
        cost_hamil = Hamiltonian(
            [
                PauliOp("ZZ", (0, 3)),
                PauliOp("ZZ", (0, 1)),
                PauliOp("ZZ", (1, 2)),
                PauliOp("ZZ", (2, 3)),
                PauliOp("Z", (0,)),
                PauliOp("Z", (1,)),
                PauliOp("Z", (2,)),
                PauliOp("Z", (3,)),
            ],
            [2.5, 2.5, 2.5, 2.5, 4.5, 4.5, 4.5, 4.5],
            12,
        )

        # From a statevector backend
        optimized_measurement_outcomes_sv = np.array(
            [
                0.18256422 - 0.1296918j,
                -0.03786405 - 0.13158363j,
                -0.03786405 - 0.13158363j,
                -0.01910622 + 0.3151214j,
                -0.03786405 - 0.13158363j,
                0.42313283 + 0.15529604j,
                -0.01910622 + 0.3151214j,
                0.016895 + 0.05117122j,
                -0.03786405 - 0.13158363j,
                -0.01910622 + 0.3151214j,
                0.42313283 + 0.15529604j,
                0.016895 + 0.05117122j,
                -0.01910622 + 0.3151214j,
                0.016895 + 0.05117122j,
                0.016895 + 0.05117122j,
                0.23799586 + 0.04022615j,
            ]
        )

        # from a shot simulator
        optimized_measurement_outcomes_shot = {
            "0010": 1,
            "0110": 2,
            "1000": 1,
            "1001": 13,
            "0000": 1,
            "1100": 10,
            "1110": 201,
            "1111": 37,
            "0101": 225,
            "1101": 57,
            "0011": 8,
            "1010": 228,
            "0111": 74,
            "1011": 115,
        }

        mps_sv = most_probable_bitstring(
            cost_hamil, qaoa_probabilities(optimized_measurement_outcomes_sv)
        )
        mps_shot = most_probable_bitstring(
            cost_hamil, optimized_measurement_outcomes_shot
        )

        assert set(["0101", "1010"]) == set(
            mps_sv["solutions_bitstrings"]
        )  # Equality for sv
        assert set(["0101", "1010"]) >= set(
            mps_shot["solutions_bitstrings"]
        )  # Subset for shot/qpu

    def test_plot_cost(self):
        # Create the problem
        g = nx.circulant_graph(6, [1])
        vc = MinimumVertexCover(g, field=1.0, penalty=10).qubo

        q = QAOA()
        q.compile(vc, verbose=False)
        q.optimize()

        q.result.plot_cost()

    def test_plot_probabilities(self):
        # Create the problem
        g = nx.circulant_graph(6, [1])
        vc = MinimumVertexCover(g, field=1.0, penalty=10).qubo

        # first for state-vector based simulator:
        q_sv = QAOA()
        q_sv.set_circuit_properties(p=3, init_type="ramp")
        q_sv.compile(vc)

        q_sv.optimize()

        q_sv.result.plot_probabilities()

    def test_plot_n_shots(self):
        # Create the problem
        g = nx.circulant_graph(6, [1])
        vc = MinimumVertexCover(g, field=1.0, penalty=10).qubo

        for method in ["cans", "icans"]:
            q = QAOA()
            q.set_classical_optimizer(
                method=method,
                maxiter=50,
                jac="finite_difference",
                optimizer_options={
                    "stepsize": 0.01,
                    "n_shots_min": 5,
                    "n_shots_max": 50,
                    "n_shots_budget": 1000,
                },
            )
            q.compile(vc, verbose=False)
            q.optimize()

            q.result.plot_n_shots()

            # all combinations to check
            test_dict = {
                "none": (
                    None,
                    {
                        "label": [None, ["t1", "t2"]],
                        "linestyle": ["-", ["--", "-"]],
                        "color": [None, ["red", "green"]],
                    },
                ),
                "int": (
                    0,
                    {
                        "label": [None, "t1", ["t2"]],
                        "linestyle": ["-", ["--"]],
                        "color": [None, ["red"], "green"],
                    },
                ),
                "list_one": (
                    1,
                    {
                        "label": [None, "t1", ["t2"]],
                        "linestyle": ["-", ["--"]],
                        "color": [None, ["red"], "green"],
                    },
                ),
                "list_two": (
                    [0, 1],
                    {
                        "label": [None, ["t1", "t2"]],
                        "linestyle": ["-", ["--", "-"]],
                        "color": [None, ["red", "green"]],
                    },
                ),
            }
            # using the test_dict we plot with different options
            for value in test_dict.values():
                for label, line, color in zip(
                    value[1]["label"], value[1]["linestyle"], value[1]["color"]
                ):
                    q.result.plot_n_shots(
                        param_to_plot=value[0],
                        label=label,
                        linestyle=line,
                        color=color,
                        title=f"method: {method}, param_to_plot: {value[0]}, label: {label}, linestyle: {line}, color: {color}",
                    )
                    plt.close()

            # function to test that errors are raised, when trying to plot with incorrect inputs
            def test_incorrect_arguments(argument: str, inputs_to_try: list):
                for x in inputs_to_try:
                    error = False
                    try:
                        q.result.plot_n_shots(**{argument: x})
                    except Exception as e:
                        assert len(str(e)) > 0, "No error message was raised"
                        error = True
                    assert error, "Error not raised, when it should have been"
                plt.close()

            # check that errors are raised, when trying to plot with incorrect inputs
            test_incorrect_arguments(
                argument="param_to_plot", inputs_to_try=["0", 2, [0, 1, 2]]
            )
            test_incorrect_arguments(
                argument="linestyle",
                inputs_to_try=[0, ["one", "two", "three"], [1, "two"]],
            )
            test_incorrect_arguments(
                argument="label", inputs_to_try=[0, ["one", "two", "three"], [1, "two"]]
            )
            test_incorrect_arguments(
                argument="color", inputs_to_try=[0, ["b", "c", "g"], [1, "g"]]
            )

    def test_get_counts(self):
        # measurement outcome from a statevector_simulator backend
        optimized_measurement_outcomes_sv = np.array(
            [
                0.18256422 - 0.1296918j,
                -0.03786405 - 0.13158363j,
                -0.03786405 - 0.13158363j,
                -0.01910622 + 0.3151214j,
                -0.03786405 - 0.13158363j,
                0.42313283 + 0.15529604j,
                -0.01910622 + 0.3151214j,
                0.016895 + 0.05117122j,
                -0.03786405 - 0.13158363j,
                -0.01910622 + 0.3151214j,
                0.42313283 + 0.15529604j,
                0.016895 + 0.05117122j,
                -0.01910622 + 0.3151214j,
                0.016895 + 0.05117122j,
                0.016895 + 0.05117122j,
                0.23799586 + 0.04022615j,
            ]
        )

        counts_from_sv = qaoa_probabilities(optimized_measurement_outcomes_sv)

        # measurement outcome from a shot simulator or QPU
        optimized_measurement_outcomes_shot = {
            "0010": 1,
            "0110": 2,
            "1000": 1,
            "1001": 13,
            "0000": 1,
            "1100": 10,
            "1110": 201,
            "1111": 37,
            "0101": 225,
            "1101": 57,
            "0011": 8,
            "1010": 228,
            "0111": 74,
            "1011": 115,
        }

        assert optimized_measurement_outcomes_shot == QAOAResult.get_counts(
            optimized_measurement_outcomes_shot
        )
        assert counts_from_sv == QAOAResult.get_counts(
            optimized_measurement_outcomes_sv
        )

    def test_best_result(self):
        """Test lowest_cost_bitstring attribute and FromDocplex2IsingModel model generation"""

        mdl = Model()  # Create a docplex model
        x = mdl.binary_var_list(5, name="x")
        mdl.maximize(np.sum(x))  # Add an objective function
        mdl.add_constraint(np.sum(x[:3]) == 0)
        q = QAOA()
        qubo = FromDocplex2IsingModel(mdl)  # translate the docplex model to qubo
        q.compile(qubo.ising_model)  # complining the ising representation of the qubo
        q.optimize()
        result = q.result
        lowest_energy = result.lowest_cost_bitstrings()
        assert (
            lowest_energy["solutions_bitstrings"][0] == "00011"
        )  # bitstring optimal solution
        assert np.isclose(lowest_energy["bitstrings_energies"][0], -2.0)  # solution


if __name__ == "__main__":
    unittest.main()



import unittest

from openqaoa.optimizers.logger_vqa import (
    Logger,
    LoggerVariable,
    LoggerVariableFactory,
    EmptyValue,
    AppendValue,
    ReplaceValue,
    IfLowerDo,
    IfHigherDo,
)


class TestingLoggerClass(unittest.TestCase):
    def test_logger_var_setting_1(self):
        test_var = LoggerVariable("test_var", EmptyValue, EmptyValue)

        test_var.update(10)

        self.assertEqual(test_var.history, [])
        self.assertEqual(test_var.best, [])

    def test_logger_var_setting_2(self):
        test_var = LoggerVariable("test_var", AppendValue, AppendValue)

        test_var.update(5)
        test_var.update(10)

        self.assertEqual(test_var.history, [5, 10])
        self.assertEqual(test_var.best, [5, 10])

    def test_logger_var_setting_3(self):
        test_var = LoggerVariable("test_var", ReplaceValue, ReplaceValue)

        test_var.update(5)
        test_var.update(10)

        self.assertEqual(test_var.history, [10])
        self.assertEqual(test_var.best, [10])

    def test_logger_var_setting_4(self):
        test_var = LoggerVariable(
            "test_var", IfLowerDo(EmptyValue), IfLowerDo(EmptyValue)
        )

        test_var.update(5)  # First value gets appended regardless
        test_var.update(10)

        self.assertEqual(test_var.history, [5])
        self.assertEqual(test_var.best, [5])

        test_var.update(1)

        self.assertEqual(test_var.history, [])
        self.assertEqual(test_var.best, [])

    def test_logger_var_setting_5(self):
        test_var = LoggerVariable(
            "test_var", IfLowerDo(AppendValue), IfLowerDo(AppendValue)
        )

        test_var.update(5)  # First value gets appended regardless
        test_var.update(10)

        self.assertEqual(test_var.history, [5])
        self.assertEqual(test_var.best, [5])

        test_var.update(1)

        self.assertEqual(test_var.history, [5, 1])
        self.assertEqual(test_var.best, [5, 1])

    def test_logger_var_setting_6(self):
        test_var = LoggerVariable(
            "test_var", IfLowerDo(ReplaceValue), IfLowerDo(ReplaceValue)
        )

        test_var.update(5)  # First value gets appended regardless
        test_var.update(10)

        self.assertEqual(test_var.history, [5])
        self.assertEqual(test_var.best, [5])

        test_var.update(1)

        self.assertEqual(test_var.history, [1])
        self.assertEqual(test_var.best, [1])

    def test_logger_var_setting_7(self):
        test_var = LoggerVariable(
            "test_var", IfHigherDo(EmptyValue), IfHigherDo(EmptyValue)
        )

        test_var.update(5)  # First value gets appended regardless
        test_var.update(1)

        self.assertEqual(test_var.history, [5])
        self.assertEqual(test_var.best, [5])

        test_var.update(10)

        self.assertEqual(test_var.history, [])
        self.assertEqual(test_var.best, [])

    def test_logger_var_setting_8(self):
        test_var = LoggerVariable(
            "test_var", IfHigherDo(AppendValue), IfHigherDo(AppendValue)
        )

        test_var.update(5)  # First value gets appended regardless
        test_var.update(1)

        self.assertEqual(test_var.history, [5])
        self.assertEqual(test_var.best, [5])

        test_var.update(10)

        self.assertEqual(test_var.history, [5, 10])
        self.assertEqual(test_var.best, [5, 10])

    def test_logger_var_setting_9(self):
        test_var = LoggerVariable(
            "test_var", IfHigherDo(ReplaceValue), IfHigherDo(ReplaceValue)
        )

        test_var.update(5)  # First value gets appended regardless
        test_var.update(1)

        self.assertEqual(test_var.history, [5])
        self.assertEqual(test_var.best, [5])

        test_var.update(10)

        self.assertEqual(test_var.history, [10])
        self.assertEqual(test_var.best, [10])

    def test_logger_var_methods(self):
        test_var = LoggerVariable("test_var", ReplaceValue, ReplaceValue)
        test_var_2 = LoggerVariable("test_var", ReplaceValue, ReplaceValue)

        test_var.update(1)

        test_var_2.update_history(1)
        test_var_2.update_best(1)

        self.assertEqual(test_var.history, test_var_2.history)
        self.assertEqual(test_var.best, test_var_2.best)

    def test_logger_var_fact_hist_bool(self):
        var_1 = LoggerVariableFactory.create_logger_variable(
            "new_attribute", True, "Replace"
        )
        var_2 = LoggerVariableFactory.create_logger_variable(
            "new_attribute", False, "Replace"
        )

        var_1.update(1)
        var_2.update(1)

        var_1.update(2)
        var_2.update(2)

        self.assertEqual(var_1.name, "new_attribute")
        self.assertEqual(var_2.name, "new_attribute")
        self.assertEqual(var_1.history, [1, 2])
        self.assertEqual(var_2.history, [])

    def test_logger_var_fact_best_str(self):
        """
        Testing all possible settings of a variable created through the Factory
        Class
        """

        var_1 = LoggerVariableFactory.create_logger_variable(
            "new_attribute", True, "Append"
        )
        var_2 = LoggerVariableFactory.create_logger_variable(
            "new_attribute", True, "Replace"
        )
        var_3 = LoggerVariableFactory.create_logger_variable(
            "new_attribute", True, "LowestSoFar"
        )
        var_4 = LoggerVariableFactory.create_logger_variable(
            "new_attribute", True, "LowestOnly"
        )
        var_5 = LoggerVariableFactory.create_logger_variable(
            "new_attribute", True, "HighestSoFar"
        )
        var_6 = LoggerVariableFactory.create_logger_variable(
            "new_attribute", True, "HighestOnly"
        )

        values_list = [5, 7, 10, 1]

        for each_value in values_list:
            var_1.update(each_value)
            var_2.update(each_value)
            var_3.update(each_value)
            var_4.update(each_value)
            var_5.update(each_value)
            var_6.update(each_value)

        self.assertEqual(var_1.best, [5, 7, 10, 1])
        self.assertEqual(var_2.best, [1])
        self.assertEqual(var_3.best, [5, 1])
        self.assertEqual(var_4.best, [1])
        self.assertEqual(var_5.best, [5, 7, 10])
        self.assertEqual(var_6.best, [10])

    def test_logger_obj_update_struct_1(self):
        """
        The best update structure ensures that the best value for that particular
        attribute respects a relation with another attribute.

        In the code below, the following can be observed:
        Even though the lowest in attribute 2 is 10. The best of attribute 2 is
        15 instead. This is because the best value of attribute 2 depends on whether
        attribute 1 for that set of update values is the highest. If it is not the
        highest value, attribute 2's best value is not updated. This way updating
        the best value of attribute 2 respects the update on the best value of
        attribute 1.

        Histories are always updated Indepedent of the best update structure provided.
        """

        logger_obj = Logger(
            {
                "attribute_1": {
                    "history_update_bool": True,
                    "best_update_string": "HighestOnly",
                },
                "attribute_2": {
                    "history_update_bool": True,
                    "best_update_string": "LowestSoFar",
                },
                "attribute_3": {
                    "history_update_bool": False,
                    "best_update_string": "Replace",
                },
            },
            {
                "root_nodes": ["attribute_1"],
                "best_update_structure": (
                    ["attribute_1", "attribute_2"],
                    ["attribute_1", "attribute_3"],
                ),
            },
        )

        logger_obj.log_variables(
            {"attribute_1": 10, "attribute_2": 3, "attribute_3": "string 1"}
        )

        self.assertEqual(logger_obj.attribute_1.history, [10])
        self.assertEqual(logger_obj.attribute_2.history, [3])
        self.assertEqual(logger_obj.attribute_3.history, [])

        self.assertEqual(logger_obj.attribute_1.best, [10])
        self.assertEqual(logger_obj.attribute_2.best, [3])
        self.assertEqual(logger_obj.attribute_3.best, ["string 1"])

        logger_obj.log_variables(
            {"attribute_1": 5, "attribute_2": 1, "attribute_3": "string 2"}
        )

        self.assertEqual(logger_obj.attribute_1.history, [10, 5])
        self.assertEqual(logger_obj.attribute_2.history, [3, 1])
        self.assertEqual(logger_obj.attribute_3.history, [])

        self.assertEqual(logger_obj.attribute_1.best, [10])
        self.assertEqual(logger_obj.attribute_2.best, [3])
        self.assertEqual(logger_obj.attribute_3.best, ["string 1"])

        logger_obj.log_variables(
            {"attribute_1": 15, "attribute_2": 4, "attribute_3": "string 3"}
        )

        self.assertEqual(logger_obj.attribute_1.history, [10, 5, 15])
        self.assertEqual(logger_obj.attribute_2.history, [3, 1, 4])
        self.assertEqual(logger_obj.attribute_3.history, [])

        self.assertEqual(logger_obj.attribute_1.best, [15])
        self.assertEqual(logger_obj.attribute_2.best, [3])
        self.assertEqual(logger_obj.attribute_3.best, ["string 3"])

        logger_obj.log_variables(
            {"attribute_1": 20, "attribute_2": 2, "attribute_3": "string 4"}
        )

        self.assertEqual(logger_obj.attribute_1.history, [10, 5, 15, 20])
        self.assertEqual(logger_obj.attribute_2.history, [3, 1, 4, 2])
        self.assertEqual(logger_obj.attribute_3.history, [])

        self.assertEqual(logger_obj.attribute_1.best, [20])
        self.assertEqual(logger_obj.attribute_2.best, [3, 2])
        self.assertEqual(logger_obj.attribute_3.best, ["string 4"])

    def test_logger_obj_update_struct_2(self):
        """
        Single layer test
        """

        logger_obj = Logger(
            {
                "attribute_1": {
                    "history_update_bool": True,
                    "best_update_string": "HighestOnly",
                },
                "attribute_2": {
                    "history_update_bool": True,
                    "best_update_string": "LowestSoFar",
                },
                "attribute_3": {
                    "history_update_bool": False,
                    "best_update_string": "Replace",
                },
            },
            {
                "root_nodes": ["attribute_1", "attribute_2", "attribute_3"],
                "best_update_structure": ([]),
            },
        )

        logger_obj.log_variables(
            {"attribute_1": 10, "attribute_2": 3, "attribute_3": "string 1"}
        )

        self.assertEqual(logger_obj.attribute_1.history, [10])
        self.assertEqual(logger_obj.attribute_2.history, [3])
        self.assertEqual(logger_obj.attribute_3.history, [])

        self.assertEqual(logger_obj.attribute_1.best, [10])
        self.assertEqual(logger_obj.attribute_2.best, [3])
        self.assertEqual(logger_obj.attribute_3.best, ["string 1"])

        logger_obj.log_variables(
            {"attribute_1": 5, "attribute_2": 1, "attribute_3": "string 2"}
        )

        self.assertEqual(logger_obj.attribute_1.history, [10, 5])
        self.assertEqual(logger_obj.attribute_2.history, [3, 1])
        self.assertEqual(logger_obj.attribute_3.history, [])

        self.assertEqual(logger_obj.attribute_1.best, [10])
        self.assertEqual(logger_obj.attribute_2.best, [3, 1])
        self.assertEqual(logger_obj.attribute_3.best, ["string 2"])

        logger_obj.log_variables(
            {"attribute_1": 15, "attribute_2": 4, "attribute_3": "string 3"}
        )

        self.assertEqual(logger_obj.attribute_1.history, [10, 5, 15])
        self.assertEqual(logger_obj.attribute_2.history, [3, 1, 4])
        self.assertEqual(logger_obj.attribute_3.history, [])

        self.assertEqual(logger_obj.attribute_1.best, [15])
        self.assertEqual(logger_obj.attribute_2.best, [3, 1])
        self.assertEqual(logger_obj.attribute_3.best, ["string 3"])

        logger_obj.log_variables(
            {"attribute_1": 20, "attribute_2": 2, "attribute_3": "string 4"}
        )

        self.assertEqual(logger_obj.attribute_1.history, [10, 5, 15, 20])
        self.assertEqual(logger_obj.attribute_2.history, [3, 1, 4, 2])
        self.assertEqual(logger_obj.attribute_3.history, [])

        self.assertEqual(logger_obj.attribute_1.best, [20])
        self.assertEqual(logger_obj.attribute_2.best, [3, 1])
        self.assertEqual(logger_obj.attribute_3.best, ["string 4"])

    def test_logger_obj_update_struct_3(self):
        """
        Multi layer test
        """

        logger_obj = Logger(
            {
                "attribute_1": {
                    "history_update_bool": True,
                    "best_update_string": "HighestOnly",
                },
                "attribute_2": {
                    "history_update_bool": True,
                    "best_update_string": "LowestSoFar",
                },
                "attribute_3": {
                    "history_update_bool": False,
                    "best_update_string": "Replace",
                },
            },
            {
                "root_nodes": ["attribute_1"],
                "best_update_structure": (
                    ["attribute_1", "attribute_2"],
                    ["attribute_2", "attribute_3"],
                ),
            },
        )

        logger_obj.log_variables(
            {"attribute_1": 10, "attribute_2": 3, "attribute_3": "string 1"}
        )

        self.assertEqual(logger_obj.attribute_1.history, [10])
        self.assertEqual(logger_obj.attribute_2.history, [3])
        self.assertEqual(logger_obj.attribute_3.history, [])

        self.assertEqual(logger_obj.attribute_1.best, [10])
        self.assertEqual(logger_obj.attribute_2.best, [3])
        self.assertEqual(logger_obj.attribute_3.best, ["string 1"])

        logger_obj.log_variables(
            {"attribute_1": 5, "attribute_2": 1, "attribute_3": "string 2"}
        )

        self.assertEqual(logger_obj.attribute_1.history, [10, 5])
        self.assertEqual(logger_obj.attribute_2.history, [3, 1])
        self.assertEqual(logger_obj.attribute_3.history, [])

        self.assertEqual(logger_obj.attribute_1.best, [10])
        self.assertEqual(logger_obj.attribute_2.best, [3])
        self.assertEqual(logger_obj.attribute_3.best, ["string 1"])

        logger_obj.log_variables(
            {"attribute_1": 15, "attribute_2": 4, "attribute_3": "string 3"}
        )

        self.assertEqual(logger_obj.attribute_1.history, [10, 5, 15])
        self.assertEqual(logger_obj.attribute_2.history, [3, 1, 4])
        self.assertEqual(logger_obj.attribute_3.history, [])

        self.assertEqual(logger_obj.attribute_1.best, [15])
        self.assertEqual(logger_obj.attribute_2.best, [3])
        self.assertEqual(logger_obj.attribute_3.best, ["string 1"])

        logger_obj.log_variables(
            {"attribute_1": 20, "attribute_2": 2, "attribute_3": "string 4"}
        )

        self.assertEqual(logger_obj.attribute_1.history, [10, 5, 15, 20])
        self.assertEqual(logger_obj.attribute_2.history, [3, 1, 4, 2])
        self.assertEqual(logger_obj.attribute_3.history, [])

        self.assertEqual(logger_obj.attribute_1.best, [20])
        self.assertEqual(logger_obj.attribute_2.best, [3, 2])
        self.assertEqual(logger_obj.attribute_3.best, ["string 4"])


if __name__ == "__main__":
    unittest.main()



# Copyright 2018-2021 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# NOTICE #
# Lines modified by Entropica Labs will bear the comment # changed from the original PennyLane code

"""Root mean square propagation optimizer"""
from numpy import sqrt  # changed from the original PennyLane code
from .adagrad import AdagradOptimizer


class RMSPropOptimizer(AdagradOptimizer):
    r"""Root mean squared propagation optimizer.

    The root mean square progation optimizer is a modified
    :class:`Adagrad optimizer <pennylane.optmimize.AdagradOptimizer>`,
    with a decay of learning rate adaptation.

    Extensions of the Adagrad optimization method generally
    start the sum :math:`a` over past gradients in the denominator
    of the learning rate at a finite :math:`t'` with :math:`0 < t' < t`,
    or decay past gradients to avoid an ever-decreasing learning rate.

    Root Mean Square propagation is such an adaptation, where

    .. math::
        a_i^{(t+1)} = \gamma a_i^{(t)} + (1-\gamma) (\partial_{x_i} f(x^{(t)}))^2.

    Args:
        stepsize (float): the user-defined hyperparameter :math:`\eta`
            used in the Adagrad optmization
        decay (float): the learning rate decay :math:`\gamma`
        eps (float): offset :math:`\epsilon` added for numerical stability
            (see :class:`Adagrad <pennylane.optmimize.AdagradOptimizer>`)

    """

    def __init__(self, stepsize=0.01, decay=0.9, eps=1e-8):
        super().__init__(stepsize)
        self.decay = decay
        self.eps = eps

    def apply_grad(self, grad, args):
        r"""Update the variables args to take a single optimization step. Flattens and unflattens
        the inputs to maintain nested iterables as the parameters of the optimization.

        Args:
            grad (tuple [array]): the gradient of the objective function at
                point :math:`x^{(t)}`: :math:`\nabla f(x^{(t)})`.
            args (tuple): the current value of the variables :math:`x^{(t)}`.

        Returns:
            list [array]: the new values :math:`x^{(t+1)}`
        """
        args_new = list(args)

        if self.accumulation is None:
            self.accumulation = [0.0] * len(args)

        trained_index = 0
        for index, arg in enumerate(args):
            if getattr(arg, "requires_grad", False):
                self._update_accumulation(index, grad[trained_index])
                args_new[index] = (
                    arg
                    - (self.stepsize / sqrt(self.accumulation[index] + self.eps))
                    * grad[trained_index]
                )
                trained_index += 1

        return args_new

    def _update_accumulation(self, index, grad):
        r"""Update the accumulation with the gradient.

        Args:
            index (int): index of argument to update.
            grad (ndarray): gradient at the index.
        """
        self.accumulation[index] = (
            self.decay * self.accumulation[index] + (1 - self.decay) * grad**2
        )



from __future__ import annotations

from .plugin_finder import PLUGIN_DICT
from .devices_core import DeviceBase, DeviceLocal


def device_class_arg_mapper(
    device_class: DeviceBase,
    hub: str = None,
    group: str = None,
    project: str = None,
    as_emulator: bool = None,
    as_qvm: bool = None,
    noisy: bool = None,
    compiler_timeout: float = None,
    execution_timeout: float = None,
    client_configuration: QCSClientConfiguration = None,
    endpoint_id: str = None,
    engagement_manager: EngagementManager = None,
    folder_name: str = None,
    s3_bucket_name: str = None,
    aws_region: str = None,
    resource_id: str = None,
    az_location: str = None,
) -> dict:
    DEVICE_ARGS_MAPPER = dict()

    local_vars = locals()

    for each_plugin_entrypoint in PLUGIN_DICT.values():
        if hasattr(each_plugin_entrypoint, "device_args"):
            for each_key, each_value in each_plugin_entrypoint.device_args.items():
                # Convert list of accepted parameters into a dictionary with
                # the name of the variable as a key and the local value of the
                # variable
                var_values = [local_vars[each_name] for each_name in each_value]
                input_dict = {each_key: dict(zip(each_value, var_values))}
                DEVICE_ARGS_MAPPER.update(input_dict)

    final_device_kwargs = {
        key: value
        for key, value in DEVICE_ARGS_MAPPER[device_class].items()
        if value is not None
    }
    return final_device_kwargs


def create_device(location: str, name: str, **kwargs):
    """
    This function returns an instance of the appropriate device class.

    Parameters
    ----------
    device_name: str
        The name of the device to be accessed.
    device_location: str
        The location of the device to be accessed.
    kwargs: dict
        A dictionary of keyword arguments to be passed to the device class.
        These will be used to initialise the device.

    Returns
    -------
    device: DeviceBase
        An instance of the appropriate device class.
    """

    location = location.lower()

    location_device_mapper = dict()
    location_device_mapper.update({"local": DeviceLocal})
    location_device_mapper.update(
        zip(
            [each_value.device_location for each_value in PLUGIN_DICT.values()],
            [each_value.device_plugin for each_value in PLUGIN_DICT.values()],
        )
    )

    if location in location_device_mapper.keys():
        device_class = location_device_mapper[location]
    else:
        raise ValueError(
            f"Invalid device location, Choose from: {location_device_mapper.keys()}"
        )

    return device_class(device_name=name, **kwargs)



from typing import List, Union
import matplotlib.pyplot as plt
from matplotlib.ticker import MaxNLocator
import numpy as np
from scipy.fftpack import dct, dst

from .variational_baseparams import QAOAVariationalBaseParams
from ..ansatz_constructor import QAOADescriptor
from ..ansatz_constructor.baseparams import shapedArray, _is_iterable_empty


class QAOAVariationalFourierParams(QAOAVariationalBaseParams):
    """
    The QAOA parameters as the sine/cosine transform of the original gammas
    and x_rotation_angles. See "Quantum Approximate Optimization Algorithm:
    Performance, Mechanism, and Implementation on Near-Term Devices"
    [https://arxiv.org/abs/1812.01041] for a detailed description.

    Parameters
    ----------
    qaoa_descriptor:
        QAOADescriptor object with circuit instructions
    q : int
        The number of coefficients for the discrete sine and cosine transforms
        below
    u : np.array
        The discrete sine transform of the ``gammas`` in
        ``StandardParams``
    v : np.array
        The discrete cosine transform of the ``betas`` in
        ``StandardParams``
    Attributes
    ----------
    q : int
        The number of coefficients for the discrete sine and cosine transforms
        below
    u : np.array
        The discrete sine transform of the ``gammas`` in
        ``StandardParams``
    v : np.array
        The discrete cosine transform of the ``betas`` in
        ``StandardParams``
    betas: np.array
        Betas to parameterize the mixer part
    gammas: np.array
        Gammas to parameterize the cost part
    """

    def __init__(
        self,
        qaoa_descriptor: QAOADescriptor,
        q: int,
        v: List[Union[int, float]],
        u: List[Union[int, float]],
    ):
        # setup reg, qubits_singles and qubits_pairs
        super().__init__(qaoa_descriptor)
        assert q is not None, f"Depth q for {type(self).__name__} must be specified"
        self.q = q
        self.v = v
        self.u = u
        self.betas = dct(self.v, n=self.p)
        self.gammas = dst(self.u, n=self.p)

    def __repr__(self):
        string = "Fourier Parameterisation:\n"
        string += "\tp: " + str(self.p) + "\n"
        string += "\tq: " + str(self.q) + "\n"
        string += "Variational Parameters:\n"
        string += "\tv: " + str(self.v) + "\n"
        string += "\tu: " + str(self.u) + "\n"
        return string

    def __len__(self):
        return 2 * self.q

    @shapedArray
    def v(self):
        return self.q

    @shapedArray
    def u(self):
        return self.q

    @property
    def mixer_1q_angles(self):
        return 2 * np.outer(self.betas, self.mixer_1q_coeffs)

    @property
    def mixer_2q_angles(self):
        return 2 * np.outer(self.betas, self.mixer_2q_coeffs)

    @property
    def cost_1q_angles(self):
        return 2 * np.outer(self.gammas, self.cost_1q_coeffs)

    @property
    def cost_2q_angles(self):
        return 2 * np.outer(self.gammas, self.cost_2q_coeffs)

    def update_from_raw(self, new_values):
        # overwrite x_rotation_angles with new ones
        self.v = np.array(new_values[0 : self.q])
        # cut x_rotation_angles from new_values
        new_values = new_values[self.q :]
        self.u = np.array(new_values[0 : self.q])
        new_values = new_values[self.q :]

        if len(new_values) != 0:
            raise RuntimeWarning(
                "Incorrect dimension specified for new_values"
                "to construct the new v's and new u's"
            )

        # update the betas and gammas too
        self.betas = dct(self.v, n=self.p)
        self.gammas = dst(self.u, n=self.p)

    def raw(self):
        raw_data = np.concatenate((self.v, self.u))
        return raw_data

    @classmethod
    def linear_ramp_from_hamiltonian(
        cls, qaoa_descriptor: QAOADescriptor, q: int, time: float = None
    ):
        """
        NOTE: rather than implement an exact linear schedule,
        this instead implements the lowest frequency component,
        i.e. a sine curve for gammas, and a cosine for betas.

        Parameters
        ----------
        qaoa_descriptor:
                        QAOADescriptor object containing information about terms,weights,register and p

        time:
            total time. Set to ``0.7*p`` if ``None`` is passed.

        Returns
        -------
        FourierParams:
            A ``FourierParams`` object with initial parameters
            corresponding to a the 0th order Fourier component
            (a sine curve for gammas, cosine for betas)

        ToDo
        ----
        Make a more informed choice of the default value for ``q``. Probably
        depending on ``n_qubits``
        """
        assert q is not None, f"Depth q for {cls.__name__} must be specified"

        # Set default time
        if time is None:
            time = 0.7 * qaoa_descriptor.p

        # fill x_rotation_angles, z_rotation_angles and zz_rotation_angles
        # Todo make this an easier expresssion
        v = np.zeros(q)
        v[0] = 0.5 * time / qaoa_descriptor.p
        u = np.copy(v)

        # wrap it all nicely in a qaoa_parameters object
        params = cls(qaoa_descriptor, q, v, u)
        return params

    @classmethod
    def random(cls, qaoa_descriptor: QAOADescriptor, q: int, seed: int = None):
        """
        Returns
        -------
        FourierParams
            randomly initialised ``FourierParams`` object
        """
        if seed is not None:
            np.random.seed(seed)

        v = np.random.uniform(0, np.pi, q)
        u = np.random.uniform(0, np.pi, q)

        params = cls(qaoa_descriptor, q, v, u)
        return params

    @classmethod
    def empty(cls, qaoa_descriptor: QAOADescriptor, q: int):
        """
        Initialise the Fourier Params u,v with empty arrays
        """
        u = np.empty((q))
        v = np.empty((q))
        return cls(qaoa_descriptor, q, v, u)

    def plot(self, ax=None, **kwargs):
        # warnings.warn("Plotting the gammas and x_rotation_angles through DCT "
        #              "and DST. If you are interested in v, u you can access "
        #              "them via params.v, params.u")

        if ax is None:
            fig, ax = plt.subplots(2, figsize=(7, 9))

        fig.tight_layout(pad=4.0)

        ax[0].plot(self.v, label="v", marker="s", ls="", **kwargs)
        ax[0].plot(self.u, label="u", marker="^", ls="", **kwargs)
        ax[0].set_xlabel("q")
        ax[0].legend()
        ax[0].xaxis.set_major_locator(MaxNLocator(integer=True))

        ax[1].plot(dct(self.v, n=self.p), label="betas", marker="s", ls="", **kwargs)
        ax[1].plot(dst(self.u, n=self.p), label="gammas", marker="^", ls="", **kwargs)
        ax[1].set_xlabel("p")
        ax[1].legend()
        ax[1].xaxis.set_major_locator(MaxNLocator(integer=True))


class QAOAVariationalFourierWithBiasParams(QAOAVariationalBaseParams):
    """
    The QAOA parameters as the sine/cosine transform of the original gammas
    and x_rotation_angles. See "Quantum Approximate Optimization Algorithm:
    Performance, Mechanism, and Implementation on Near-Term Devices"
    [https://arxiv.org/abs/1812.01041] for a detailed description.

    Parameters
    ----------
    qaoa_descriptor:
        QAOADescriptor object with circuit instructions
    q : int
        The number of coefficients for the discrete sine and cosine transforms
        below
    u_pairs : np.array
        The discrete sine transform of the ``gammas_pairs`` in
        ``StandardWithBiasParams``
    u_singles : np.array
        The discrete sine transform of the ``gammas_singles`` in
        ``StandardWithBiasParams``
    v : np.array
        The discrete cosine transform of the betas in
        ``StandardWithBiasParams``

    Attributes
    ----------
    q : int
        The number of coefficients for the discrete sine and cosine transforms
        below
    u_pairs : np.array
        The discrete sine transform of the ``gammas_pairs`` in
        ``StandardWithBiasParams``
    u_singles : np.array
        The discrete sine transform of the ``gammas_singles`` in
        ``StandardWithBiasParams``
    v : np.array
        The discrete cosine transform of the betas in
        ``StandardWithBiasParams``
    betas: np.array
    gammas_singles: np.array
    gammas_pairs: np.array
    """

    def __init__(
        self,
        qaoa_descriptor: QAOADescriptor,
        q: int,
        v: List[Union[float, int]],
        u_singles: List[Union[float, int]],
        u_pairs: List[Union[float, int]],
    ):
        # setup reg, qubits_singles and qubits_pairs
        super().__init__(qaoa_descriptor)
        if not self.cost_1q_coeffs or not self.cost_2q_coeffs:
            raise RuntimeError(
                f"Please choose {type(self).__name__} parameterisation for problems "
                "containing both Cost One-Qubit and Two-Qubit terms"
            )
        assert q is not None, f"Depth q for {type(self).__name__} must be specified"

        self.q = q
        self.v = v
        self.u_singles = u_singles
        self.u_pairs = u_pairs

        self.betas = dct(self.v, n=self.p)
        self.gammas_singles = dst(self.u_singles, n=self.p)
        self.gammas_pairs = dst(self.u_pairs, n=self.p)

    def __repr__(self):
        string = "Fourier Parameterisation:\n"
        string += "\tp: " + str(self.p) + "\n"
        string += "\tq: " + str(self.q) + "\n"
        string += "Variational Parameters:\n"
        string += "\tv: " + str(self.v) + "\n"
        string += "\tu_singles: " + str(self.u_singles) + "\n"
        string += "\tu_pairs: " + str(self.u_pairs) + "\n"
        return string

    def __len__(self):
        return 3 * self.q

    @shapedArray
    def v(self):
        return self.q

    @shapedArray
    def u_singles(self):
        return self.q

    @shapedArray
    def u_pairs(self):
        return self.q

    @property
    def mixer_1q_angles(self):
        return 2 * np.outer(self.betas, self.mixer_1q_coeffs)

    @property
    def mixer_2q_angles(self):
        return 2 * np.outer(self.betas, self.mixer_2q_coeffs)

    @property
    def cost_1q_angles(self):
        return 2 * np.outer(self.gammas_singles, self.cost_1q_coeffs)

    @property
    def cost_2q_angles(self):
        return 2 * np.outer(self.gammas_pairs, self.cost_2q_coeffs)

    def update_from_raw(self, new_values):
        # overwrite x_rotation_angles with new ones
        self.v = np.array(new_values[0 : self.q])
        # cut x_rotation_angles from new_values
        new_values = new_values[self.q :]

        self.u_singles = np.array(new_values[0 : self.q])
        new_values = new_values[self.q :]

        self.u_pairs = np.array(new_values[0 : self.q])
        new_values = new_values[self.q :]

        if len(new_values) != 0:
            raise RuntimeWarning(
                "Incorrect dimension specified for new_values"
                "to construct the new v's and new u's"
            )

        # update the betas, gammas too
        self.betas = dct(self.v, n=self.p)
        self.gammas_singles = dst(self.u_singles, n=self.p)
        self.gammas_pairs = dst(self.u_pairs, n=self.p)

    def raw(self):
        raw_data = np.concatenate((self.v, self.u_singles, self.u_pairs))
        return raw_data

    @classmethod
    def linear_ramp_from_hamiltonian(
        cls, qaoa_descriptor: QAOADescriptor, q: int, time: float = None
    ):
        """
        Parameters
        ----------
        qaoa_descriptor:
            Hyper Parameters passed as an object of ``Type[QAOADescriptor]``

        time:
            Time for creating the linear ramp schedule.
            Defaults to ``0.7*p`` if None

        Returns
        -------
        FourierWithBiasParams:
            A ``FourierWithBiasParams`` object with initial parameters
            corresponding to a linear ramp annealing schedule

        ToDo
        ----
        Make a more informed choice of the default value for ``q``. Probably
        depending on ``n_qubits``
        """

        assert q is not None, f"Depth q for {cls.__name__} must be specified"

        if time is None:
            time = 0.7 * qaoa_descriptor.p

        v = np.zeros(q)
        v[0] = 0.5 * time / qaoa_descriptor.p
        u_singles = np.copy(v)
        u_pairs = np.copy(v)
        # wrap it all nicely in a qaoa_parameters object
        params = cls(qaoa_descriptor, q, v, u_singles, u_pairs)
        return params

    @classmethod
    def random(cls, qaoa_descriptor: QAOADescriptor, q: int, seed: int = None):
        """
        Returns
        -------
        FourierWithBiasParams
            randomly initialised ``FourierWithBiasParams`` object
        """
        if seed is not None:
            np.random.seed(seed)

        v = np.random.uniform(0, np.pi, q)
        u_singles = np.random.uniform(0, np.pi, q)
        u_pairs = np.random.uniform(0, np.pi, q)

        params = cls(qaoa_descriptor, q, v, u_singles, u_pairs)
        return params

    @classmethod
    def empty(cls, qaoa_descriptor: QAOADescriptor, q: int):
        """
        Initialise Fourier parameters with bias with an empty array
        """
        v = np.empty((q))
        u_singles = np.empty((q))
        u_pairs = np.empty((q))
        return cls(qaoa_descriptor, q, v, u_singles, u_pairs)

    def plot(self, ax=None, **kwargs):
        # warnings.warn("Plotting the gammas and x_rotation_angles through DCT "
        #              "and DST. If you are interested in v, u_singles and "
        #              "u_pairs you can access them via params.v, "
        #              "params.u_singles, params.u_pairs")
        if ax is None:
            fig, ax = plt.subplots(2, figsize=(7, 9))

        fig.tight_layout(pad=4.0)

        ax[0].plot(self.v, label="v", marker="s", ls="", **kwargs)
        ax[0].plot(self.u_singles, label="u_singles", marker="^", ls="", **kwargs)
        ax[0].plot(self.u_pairs, label="u_pairs", marker="v", ls="", **kwargs)
        ax[0].set_xlabel("q")
        ax[0].legend()
        ax[0].xaxis.set_major_locator(MaxNLocator(integer=True))

        ax[1].plot(dct(self.v, n=self.p), label="betas", marker="s", ls="", **kwargs)
        if not _is_iterable_empty(self.u_singles):
            ax[1].plot(
                dst(self.u_singles, n=self.p),
                label="gammas_singles",
                marker="^",
                ls="",
                **kwargs,
            )
        if not _is_iterable_empty(self.u_pairs):
            ax[1].plot(
                dst(self.u_pairs, n=self.p),
                label="gammas_pairs",
                marker="v",
                ls="",
                **kwargs,
            )
        ax[1].set_xlabel("p")
        ax[1].legend()
        ax[1].xaxis.set_major_locator(MaxNLocator(integer=True))


class QAOAVariationalFourierExtendedParams(QAOAVariationalBaseParams):
    r"""
    The Fourier pendant to ExtendedParams.

    Parameters
    ----------
    qaoa_descriptor: ``QAOADescriptor``
        object containing information about terms,weights,register and p
    q: ``int``
        The parameter depth for u and v Fourier params
    v: np.array
        The discrete cosine transform of the ``betas`` in ``ExtendedParams``
    u_singles: np.array
        The discrete sine transform of the ``gammas_singles`` in
        ``ExtendedParams``
    u_pairs: np.array
        The discrete sine transform of the ``gammas_pairs`` in
        ``ExtendedParams``

    Attributes
    ----------
    q : int
        The number of coefficients for the discrete sine and cosine transforms
        below
    v: np.array
        The discrete cosine transform of the ``betas`` in ``ExtendedParams``
    u_singles: np.array
        The discrete sine transform of the ``gammas_singles`` in
        ``ExtendedParams``
    u_pairs: np.array
        The discrete sine transform of the ``gammas_pairs`` in
        ``ExtendedParams``
    betas_singles:
    betas_pairs:
    gammas_singles:
    gammas_pairs:
    """

    def __init__(
        self,
        qaoa_descriptor: QAOADescriptor,
        q: int,
        v_singles: List[Union[float, int]],
        v_pairs: List[Union[float, int]],
        u_singles: List[Union[float, int]],
        u_pairs: List[Union[float, int]],
    ):
        # setup reg, qubits_singles and qubits_pairs
        super().__init__(qaoa_descriptor)
        assert q is not None, f"Depth q for {type(self).__name__} must be specified"
        self.q = q
        self.v_singles = v_singles
        self.v_pairs = v_pairs
        self.u_singles = u_singles
        self.u_pairs = u_pairs

        self.betas_singles = dct(self.v_singles, n=self.p, axis=0)
        self.betas_pairs = dct(self.v_pairs, n=self.p, axis=0)
        self.gammas_singles = dst(self.u_singles, n=self.p, axis=0)
        self.gammas_pairs = dst(self.u_pairs, n=self.p, axis=0)

    def __repr__(self):
        string = "Fourier Parameterisation:\n"
        string += "\tp: " + str(self.p) + "\n"
        string += "\tq: " + str(self.q) + "\n"
        string += "Variational Parameters:\n"
        string += "\tv_singles: " + str(self.v_singles).replace("\n", ",") + "\n"
        string += "\tv_pairs: " + str(self.v_pairs).replace("\n", ",") + "\n"
        string += "\tu_singles: " + str(self.u_singles) + "\n"
        string += "\tu_pairs: " + str(self.u_pairs) + "\n"
        return string

    def __len__(self):
        return self.q * (
            len(self.mixer_1q_coeffs)
            + len(self.mixer_2q_coeffs)
            + len(self.cost_1q_coeffs)
            + len(self.cost_2q_coeffs)
        )

    @shapedArray
    def v_singles(self):
        return (self.q, len(self.mixer_1q_coeffs))

    @shapedArray
    def v_pairs(self):
        return (self.q, len(self.mixer_2q_coeffs))

    @shapedArray
    def u_singles(self):
        return (self.q, len(self.cost_1q_coeffs))

    @shapedArray
    def u_pairs(self):
        return (self.q, len(self.cost_2q_coeffs))

    @property
    def mixer_1q_angles(self):
        return 2 * (self.mixer_1q_coeffs * self.betas_singles)

    @property
    def mixer_2q_angles(self):
        return 2 * (self.mixer_2q_coeffs * self.betas_pairs)

    @property
    def cost_1q_angles(self):
        if self.u_singles.size > 0:
            return 2 * (self.cost_1q_coeffs * self.gammas_singles)
        else:
            return 2 * (self.cost_1q_coeffs * np.empty(shape=(self.p, 0)))

    @property
    def cost_2q_angles(self):
        if self.u_pairs.size > 0:
            return 2 * (self.cost_2q_coeffs * self.gammas_pairs)
        else:
            return 2 * (self.cost_2q_coeffs * np.empty(shape=(self.p, 0)))

    def update_from_raw(self, new_values):
        self.v_singles = np.array(new_values[: len(self.mixer_1q_coeffs) * self.q])
        self.v_singles = self.v_singles.reshape((self.q, len(self.mixer_1q_coeffs)))
        new_values = new_values[self.q * len(self.mixer_1q_coeffs) :]

        self.v_pairs = np.array(new_values[: len(self.mixer_2q_coeffs) * self.q])
        self.v_pairs = self.v_pairs.reshape((self.q, len(self.mixer_2q_coeffs)))
        new_values = new_values[self.q * len(self.mixer_2q_coeffs) :]

        self.u_singles = np.array(new_values[: len(self.cost_1q_coeffs) * self.q])
        self.u_singles = self.u_singles.reshape((self.q, len(self.cost_1q_coeffs)))
        new_values = new_values[self.q * len(self.cost_1q_coeffs) :]

        self.u_pairs = np.array(new_values[: len(self.cost_2q_coeffs) * self.q])
        self.u_pairs = self.u_pairs.reshape((self.q, len(self.cost_2q_coeffs)))
        new_values = new_values[self.q * len(self.cost_2q_coeffs) :]

        # PEP8 complains, but new_values could be np.array and not list!
        if len(new_values) != 0:
            raise RuntimeWarning(
                "Incorrect dimension specified for new_values"
                "to construct the new v's and new u's"
            )

        # update the betas, gammas too
        self.betas_singles = dct(self.v_singles, n=self.p, axis=0)
        self.betas_pairs = dct(self.v_pairs, n=self.p, axis=0)
        self.gammas_singles = dst(self.u_singles, n=self.p, axis=0)
        self.gammas_pairs = dst(self.u_pairs, n=self.p, axis=0)

    def raw(self):
        raw_data = np.concatenate(
            (
                self.v_singles.flatten(),
                self.v_pairs.flatten(),
                self.u_singles.flatten(),
                self.u_pairs.flatten(),
            )
        )
        return raw_data

    @classmethod
    def linear_ramp_from_hamiltonian(
        cls, qaoa_descriptor: QAOADescriptor, q: int, time: float = None
    ):
        """
        Parameters
        ----------
        qaoa_descriptor: ``QAOADescriptor``
            an object containing information about the QAOA circuit

        q: ``int``
            The q-depth of the circuit parameters

        time: ``float``
            Time for creating the linear ramp schedule.
            Defaults to ``0.7*p`` if None

        Returns
        -------
        FourierExtendedParams
            The initial parameters according to a linear ramp for
            for the Hamiltonian specified by register, terms, weights.

        """

        assert q is not None, f"Depth q for {cls.__name__} must be specified"

        # create evenly spaced timelayers at the centers of p intervals
        p = qaoa_descriptor.p

        if time is None:
            time = float(0.7 * p)

        n_u_singles = len(qaoa_descriptor.cost_single_qubit_coeffs)
        n_u_pairs = len(qaoa_descriptor.cost_pair_qubit_coeffs)
        n_v_singles = len(qaoa_descriptor.mixer_single_qubit_coeffs)
        n_v_pairs = len(qaoa_descriptor.mixer_pair_qubit_coeffs)

        v = np.zeros(q)
        v[0] = 0.5 * time / p
        u = np.copy(v)

        v_singles = v.repeat(n_v_singles).reshape(q, n_v_singles)
        v_pairs = v.repeat(n_v_pairs).reshape(q, n_v_pairs)
        u_singles = u.repeat(n_u_singles).reshape(q, n_u_singles)
        u_pairs = u.repeat(n_u_pairs).reshape(q, n_u_pairs)

        # wrap it all nicely in a qaoa_parameters object
        params = cls(qaoa_descriptor, q, v_singles, v_pairs, u_singles, u_pairs)
        return params

    @classmethod
    def random(cls, qaoa_descriptor: QAOADescriptor, q: int, seed: int = None):
        """
        Returns
        -------
        FourierExtendedParams
            randomly initialised ``FourierExtendedParams`` object
        """
        if seed is not None:
            np.random.seed(seed)

        p = qaoa_descriptor.p
        n_u_singles = len(qaoa_descriptor.cost_single_qubit_coeffs)
        n_u_pairs = len(qaoa_descriptor.cost_pair_qubit_coeffs)
        n_v_singles = len(qaoa_descriptor.mixer_single_qubit_coeffs)
        n_v_pairs = len(qaoa_descriptor.mixer_pair_qubit_coeffs)

        v_singles = np.random.uniform(0, np.pi, (q, n_v_singles))
        v_pairs = np.random.uniform(0, np.pi, (q, n_v_pairs))
        u_singles = np.random.uniform(0, np.pi, (q, n_u_singles))
        u_pairs = np.random.uniform(0, np.pi, (q, n_u_pairs))

        params = cls(qaoa_descriptor, q, v_singles, v_pairs, u_singles, u_pairs)
        return params

    @classmethod
    def empty(cls, qaoa_descriptor: QAOADescriptor, q: int):
        """
        Initialise Fourier extended parameters with empty arrays
        """

        v_singles = np.empty((q, len(qaoa_descriptor.mixer_single_qubit_coeffs)))
        v_pairs = np.empty((q, len(qaoa_descriptor.mixer_pair_qubit_coeffs)))
        u_singles = np.empty((q, len(qaoa_descriptor.cost_single_qubit_coeffs)))
        u_pairs = np.empty((q, len(qaoa_descriptor.cost_pair_qubit_coeffs)))

        return cls(qaoa_descriptor, q, v_singles, v_pairs, u_singles, u_pairs)

    def plot(self, ax=None, **kwargs):
        # if ax is None:
        #     fig, ax = plt.subplots()

        # ax.plot(dct(self.v, n=self.p, axis=0),
        #         label="betas", marker="s", ls="", **kwargs)
        # if not _is_iterable_empty(self.u_singles):
        #     ax.plot(dst(self.u_singles, n=self.p),
        #             label="gammas_singles", marker="^", ls="", **kwargs)
        # if not _is_iterable_empty(self.u_pairs):
        #     ax.plot(dst(self.u_pairs, n=self.p),
        #             label="gammas_pairs", marker="v", ls="", **kwargs)
        # ax.set_xlabel("timestep")
        # ax.legend()

        betas_singles = dct(self.v_singles, n=self.p, axis=0)
        betas_pairs = dct(self.v_pairs, n=self.p, axis=0)
        gammas_singles = dst(self.u_singles, n=self.p, axis=0)
        gammas_pairs = dst(self.u_pairs, n=self.p, axis=0)

        p = self.p
        q = self.q
        list_pq_ = [q, q, q, q, p, p, p, p]
        list_pq_names_ = ["q", "q", "q", "q", "p", "p", "p", "p"]

        list_names_ = [
            "v singles",
            "v pairs",
            "u singles",
            "u pairs",
            "betas singles",
            "betas pairs",
            "gammas singles",
            "gammas pairs",
        ]
        list_values_ = [
            self.v_singles % (2 * (np.pi)),
            self.v_pairs % (2 * (np.pi)),
            self.u_singles % (2 * (np.pi)),
            self.u_pairs % (2 * (np.pi)),
            betas_singles % (2 * (np.pi)),
            betas_pairs % (2 * (np.pi)),
            gammas_singles % (2 * (np.pi)),
            gammas_pairs % (2 * (np.pi)),
        ]

        list_names, list_values = list_names_.copy(), list_values_.copy()
        list_pq, list_pq_names = list_pq_.copy(), list_pq_names_.copy()

        n_pop = 0
        for i in range(len(list_values_)):
            if list_values_[i].size == 0:
                list_values.pop(i - n_pop)
                list_names.pop(i - n_pop)
                list_pq.pop(i - n_pop)
                list_pq_names.pop(i - n_pop)
                n_pop += 1

        n = len(list_values)

        if ax is None:
            fig, ax = plt.subplots((n + 1) // 2, 2, figsize=(9, 4 * (n + 1) // 2))

        fig.tight_layout(pad=4.0)

        for k, (name, values) in enumerate(zip(list_names, list_values)):
            i, j = k // 2, k % 2
            pq = list_pq[k]
            pq_name = list_pq_names[k]
            axes = ax[i, j] if n > 2 else ax[k]

            if values.size == pq:
                axes.plot(values.T[0], marker="^", color="green", ls="", **kwargs)
                axes.set_xlabel(pq_name, fontsize=12)
                axes.set_title(name)
                axes.xaxis.set_major_locator(MaxNLocator(integer=True))

            elif values.size > 0:
                n_terms = values.shape[1]
                plt1 = axes.pcolor(
                    np.arange(pq),
                    np.arange(n_terms),
                    values.T,
                    vmin=0,
                    vmax=2 * np.pi,
                    cmap="seismic",
                )
                axes.set_aspect(pq / n_terms)
                axes.xaxis.set_major_locator(MaxNLocator(integer=True))
                axes.yaxis.set_major_locator(MaxNLocator(integer=True))
                axes.set_ylabel("terms")
                axes.set_xlabel(pq_name)
                axes.set_title(name)

                plt.colorbar(plt1, **kwargs)

        if j == 0:
            ax[i, j + 1].axis("off")



import unittest
import numpy as np
import json
import os
import pytest
import subprocess

from openqaoa.backends.qaoa_backend import (
    get_qaoa_backend,
    DEVICE_NAME_TO_OBJECT_MAPPER,
    DEVICE_ACCESS_OBJECT_MAPPER,
)
from openqaoa.qaoa_components import (
    Hamiltonian,
    create_qaoa_variational_params,
    QAOADescriptor,
)

from openqaoa_qiskit.backends import (
    QAOAQiskitBackendShotBasedSimulator,
)

from openqaoa.backends.wrapper import SPAMTwirlingWrapper

from openqaoa.utilities import X_mixer_hamiltonian, bitstring_energy
from openqaoa.backends.qaoa_device import create_device
from openqaoa.backends.basebackend import QAOABaseBackendShotBased


def get_params():
    cost_hamil = Hamiltonian.classical_hamiltonian(
        terms=[[0, 1], [0], [1]], coeffs=[1, 1, 1], constant=0
    )
    mixer_hamil = X_mixer_hamiltonian(2)

    qaoa_descriptor = QAOADescriptor(cost_hamil, mixer_hamil, p=1)
    variational_params_std = create_qaoa_variational_params(
        qaoa_descriptor, "standard", "ramp"
    )

    return qaoa_descriptor, variational_params_std


class TestingSPAMTwirlingWrapper(unittest.TestCase):
    """
    These tests check methods of the SPAM Twirling wrapper.
    """

    def setUp(
        self,
    ):
        self.n_batches = 5
        self.calibration_data_location = (
            "./tests/qpu_calibration_data/spam_twirling_mock.json"
        )
        self.qaoa_descriptor, self.variate_params = get_params()

        self.qiskit_shot_backend = QAOAQiskitBackendShotBasedSimulator(
            qaoa_descriptor=self.qaoa_descriptor,
            n_shots=100,
            prepend_state=None,
            append_state=None,
            init_hadamard=True,
            cvar_alpha=1.0,
            qiskit_simulation_method="automatic",
            seed_simulator=2642,
            noise_model=None,
        )

        self.wrapped_obj = SPAMTwirlingWrapper(
            self.qiskit_shot_backend,
            n_batches=self.n_batches,
            calibration_data_location=self.calibration_data_location,
        )

        self.wrapped_obj_trivial = SPAMTwirlingWrapper(
            self.qiskit_shot_backend,
            n_batches=1,
            calibration_data_location=self.calibration_data_location,
        )
    
    @pytest.mark.qpu
    def test_wrap_any_backend(self):
        """
        Testing if the wrapper is backend-agnostic by checking if it can take
        any of the relevant backend objects as an argument.
        """

        rigetti_args = {
            "as_qvm": True,
            "execution_timeout": 10,
            "compiler_timeout": 100,
        }
        device_list = []
        device_list.append(
            create_device(location="local", name="qiskit.qasm_simulator")
        )
        device_list.append(
            create_device(location="ibmq", name="ibm_perth"),
        )

        for device in device_list:
            backend = get_qaoa_backend(
                qaoa_descriptor=self.qaoa_descriptor,
                device=device,
                n_shots=42,
            )
            try:
                SPAMTwirlingWrapper(
                    backend,
                    n_batches=self.n_batches,
                    calibration_data_location=self.calibration_data_location,
                )
            except Exception as e:
                raise ValueError("The {} backend cannot be wrapped.".format(backend), e)

    def test_setUp(self):
        assert (
            self.wrapped_obj.n_batches == self.n_batches
        ), "The number of batches hasn't been set correctly."
        assert (
            self.wrapped_obj.calibration_data_location == self.calibration_data_location
        ), "The location of the calibration file hasn't been set correctly."

    def test_get_counts(self):
        """
        Testing the get_counts method of the SPAM Twirling wrapper.

        The test compares the probability dictionary returned from the
        get_counts method as overriden by the wrapper to the hardcoded one and
        the unwrapped backend. The later is only true for the very trivial case
        of a using a single batch and seed 1 which translates to not adding
        any X gates to the circuit. The test also compares the outcome
        probabilitites obtained within the wrapper to the ones obtained
        form the original backend. These should be almost equal given that
        there's no measurement noise and enough shots.
        """
        try:
            self.wrapped_obj
            self.wrapped_obj_trivial
        except NameError:
            print("Skipping Qiskit tests as Qiskit is not installed.")
            return None

        assert self.wrapped_obj.get_counts(
            self.variate_params, n_shots=100, seed=13
        ) == {
            "11": 27,
            "10": 36,
            "01": 35,
            "00": 2,
        }, "The get_counts function in the wrapper didn't return the expected counts."

        assert self.wrapped_obj_trivial.get_counts(
            self.variate_params, n_shots=100, seed=1
        ) == self.qiskit_shot_backend.get_counts(
            self.variate_params, n_shots=100
        ), "The get_counts function in the trivial wrapper didn't return the original backend counts."

        # Obtain counts for the trivial case with n_batches=1 but seed=1 which means applying X gates to both qubits
        n_shots = 10**6  # more counts for accurate statistics
        wrapper_counts = self.wrapped_obj_trivial.get_counts(
            self.variate_params, n_shots=n_shots, seed=2
        )
        backend_counts = self.qiskit_shot_backend.get_counts(
            self.variate_params, n_shots=n_shots
        )

        # convert counts to probabilities
        wrapper_probs = {key: value / n_shots for key, value in wrapper_counts.items()}
        backend_probs = {key: value / n_shots for key, value in backend_counts.items()}

        for key in wrapper_probs:
            assert np.isclose(
                wrapper_probs[key], backend_probs[key], rtol=1e-1
            ), "The get_counts function in the trivial wrapper when negating both qubits returned a value too far from the original backend."

        # Obtain counts for the trivial case with n_batches=5 and seed=26 which means applying X gates at random at every batch
        n_shots = 10**6  # more counts for accurate statistics
        wrapper_counts = self.wrapped_obj.get_counts(
            self.variate_params, n_shots=n_shots, seed=26
        )
        backend_counts = self.qiskit_shot_backend.get_counts(
            self.variate_params, n_shots=n_shots
        )

        # convert counts to probabilities
        wrapper_probs = {key: value / n_shots for key, value in wrapper_counts.items()}
        backend_probs = {key: value / n_shots for key, value in backend_counts.items()}

        for key in wrapper_probs:
            assert np.isclose(
                wrapper_probs[key], backend_probs[key], rtol=1e-1
            ), "The get_counts function in the wrapper when negating random qubits returned a value too far from the original backend."

    def test_expectation_value_spam_twirled(self):
        """
        Testing the expectation_value_spam_twirled method of the SPAM Twirling
        wrapper in the following scenarious:
        Given very trivial counts where only the 00 state is present and
        calibration factors 1.0, meaning no errors, the expectation value
        of the energy must be the energy of the 00 bitstring.
        Given the same counts but calibration factors 0.5, the expectation
        value of the energy must be twice the energy of the 00 bitstring
        due to corrections coming from the calibration factors.

        """
        try:
            self.wrapped_obj
            self.wrapped_obj_trivial
        except NameError:
            print("Skipping Qiskit tests as Qiskit is not installed.")
            return None

        counts = {"00": 100, "01": 0, "10": 0, "11": 0}
        hamiltonian = Hamiltonian.classical_hamiltonian(
            terms=[[0, 1], [0], [1]], coeffs=[1, 1, 1], constant=0
        )
        calibration_factors = {(1,): 1.0, (0,): 1.0, (0, 1): 1.0}
        assert self.wrapped_obj.expectation_value_spam_twirled(
            counts, hamiltonian, calibration_factors
        ) == bitstring_energy(
            hamiltonian, "00"
        ), "The function which computes the expectation value when using spam twirling with trivial calibration factors doesn't give the correct energy."

        calibration_factors = {(1,): 0.5, (0,): 0.5, (0, 1): 0.5}
        assert self.wrapped_obj.expectation_value_spam_twirled(
            counts, hamiltonian, calibration_factors
        ) == 2 * bitstring_energy(
            hamiltonian, "00"
        ), "The function which computes the expectation value when using spam twirling with calibration factors = 0.5 doesn't give the correct energy."

        counts = {"00": 25, "01": 25, "10": 25, "11": 25}
        calibration_factors = {(1,): 1.0, (0,): 1.0, (0, 1): 1.0}
        assert (
            self.wrapped_obj.expectation_value_spam_twirled(
                counts, hamiltonian, calibration_factors
            )
            == 0
        ), "The function which computes the expectation value when using spam twirling when equal superposition of states doesn't give the correct energy."

        counts = {"00": 13, "01": 26, "10": 42, "11": 66}
        calibration_factors = {(1,): 0.9, (0,): 0.8, (0, 1): 0.7}
        assert (
            self.wrapped_obj.expectation_value_spam_twirled(
                counts, hamiltonian, calibration_factors
            )
            == -0.759502213584
        ), "The function which computes the expectation value when using spam twirling with random counts and factors doesn't give the correct energy."

    def test_expectation(self):
        """
        Testing the expectation method of the SPAM Twirling wrapper which overrides the backend function as defined in basebackends.
        """
        try:
            self.wrapped_obj
            self.wrapped_obj_trivial
        except NameError:
            print("Skipping Qiskit tests as Qiskit is not installed.")
            return None

        self.assertAlmostEqual(
            self.wrapped_obj.expectation(self.variate_params, n_shots=10000), -0.94, 1
        )


if __name__ == "__main__":
    unittest.main()



# Copyright 2018-2021 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Multiple dispatch functions"""
# pylint: disable=import-outside-toplevel,too-many-return-statements
import warnings
from collections.abc import Sequence
import functools

from autograd.numpy.numpy_boxes import ArrayBox
from autoray import numpy as np
from numpy import ndarray

from . import single_dispatch  # pylint:disable=unused-import
from .utils import cast, get_interface, requires_grad


# pylint:disable=redefined-outer-name
def array(*args, like=None, **kwargs):
    """Creates an array or tensor object of the target framework.

    This method preserves the Torch device used.

    Returns:
        tensor_like: the tensor_like object of the framework
    """
    res = np.array(*args, like=like, **kwargs)
    if like is not None and get_interface(like) == "torch":
        res = res.to(device=like.device)
    return res


def eye(*args, like=None, **kwargs):
    """Creates an identity array or tensor object of the target framework.

    This method preserves the Torch device used.

    Returns:
        tensor_like: the tensor_like object of the framework
    """
    res = np.eye(*args, like=like, **kwargs)
    if like is not None and get_interface(like) == "torch":
        res = res.to(device=like.device)
    return res


def _multi_dispatch(values):
    """Determines the correct framework to dispatch to given a
    sequence of tensor-like objects.

    Args:
        values (Sequence[tensor_like]): a sequence of tensor like objects

    Returns:
        str: the name of the interface

    To determine the framework to dispatch to, the following rules
    are applied:

    * Tensors that are incompatible (such as Torch and TensorFlow tensors)
      cannot both be present.

    * Autograd tensors *may* be present alongside Torch and TensorFlow tensors,
      but Torch and TensorFlow take precendence; the autograd arrays will
      be treated as non-differentiable NumPy arrays. A warning will be raised
      suggesting that vanilla NumPy be used instead.

    * Vanilla NumPy arrays and SciPy sparse matrices can be used alongside other tensor objects;
      they will always be treated as non-differentiable constants.
    """
    if "resource_variable" in getattr(values, "__module__", tuple()):
        values = np.asarray(values)

    interfaces = {get_interface(v) for v in values}

    if len(set(interfaces) - {"numpy", "scipy", "autograd"}) > 1:
        # contains multiple non-autograd interfaces
        raise ValueError(
            "Tensors contain mixed types; cannot determine dispatch library"
        )

    non_numpy_scipy_interfaces = set(interfaces) - {"numpy", "scipy"}

    if len(non_numpy_scipy_interfaces) > 1:
        # contains autograd and another interface
        warnings.warn(
            f"Contains tensors of types {non_numpy_scipy_interfaces}; dispatch will prioritize "
            "TensorFlow and PyTorch over autograd. Consider replacing Autograd with vanilla NumPy.",
            UserWarning,
        )

    if "tensorflow" in interfaces:
        return "tensorflow"

    if "torch" in interfaces:
        return "torch"

    if "autograd" in interfaces:
        return "autograd"

    if "jax" in interfaces:
        return "jax"

    return "numpy"


def multi_dispatch(argnum=None, tensor_list=None):
    r"""Decorater to dispatch arguments handled by the interface.

    This helps simplify definitions of new functions inside PennyLane. We can
    decorate the function, indicating the arguments that are tensors handled
    by the interface:


    >>> @qml.math.multi_dispatch(argnum=[0, 1])
    ... def some_function(tensor1, tensor2, option, like):
    ...     # the interface string is stored in `like`.
    ...     ...


    Args:
        argnum (list[int]): A list of integers indicating indicating the indices
            to dispatch (i.e., the arguments that are tensors handled by an interface).
            If ``None``, dispatch over all arguments.
        tensor_lists (list[int]): a list of integers indicating which indices
            in ``argnum`` are expected to be lists of tensors. If an argument
            marked as tensor list is not a ``tuple`` or ``list``, it is treated
            as if it was not marked as tensor list. If ``None``, this option is ignored.

    Returns:
        func: A wrapped version of the function, which will automatically attempt
        to dispatch to the correct autodifferentiation framework for the requested
        arguments. Note that the ``like`` argument will be optional, but can be provided
        if an explicit override is needed.

    .. seealso:: :func:`pennylane.math.multi_dispatch._multi_dispatch`

    .. note::
        This decorator makes the interface argument "like" optional as it utilizes
        the utility function `_multi_dispatch` to automatically detect the appropriate
        interface based on the tensor types.

    **Examples**

    We can redefine external functions to be suitable for PennyLane. Here, we
    redefine Autoray's ``stack`` function.

    >>> stack = multi_dispatch(argnum=0, tensor_list=0)(autoray.numpy.stack)

    We can also use the ``multi_dispatch`` decorator to dispatch
    arguments of more more elaborate custom functions. Here is an example
    of a ``custom_function`` that
    computes :math:`c \\sum_i (v_i)^T v_i`, where :math:`v_i` are vectors in ``values`` and
    :math:`c` is a fixed ``coefficient``. Note how ``argnum=0`` only points to the first argument ``values``,
    how ``tensor_list=0`` indicates that said first argument is a list of vectors, and that ``coefficient`` is not
    dispatched.

    >>> @math.multi_dispatch(argnum=0, tensor_list=0)
    >>> def custom_function(values, like, coefficient=10):
    >>>     # values is a list of vectors
    >>>     # like can force the interface (optional)
    >>>     if like == "tensorflow":
    >>>         # add interface-specific handling if necessary
    >>>     return coefficient * np.sum([math.dot(v,v) for v in values])

    We can then run

    >>> values = [np.array([1, 2, 3]) for _ in range(5)]
    >>> custom_function(values)
    700

    """

    def decorator(fn):
        @functools.wraps(fn)
        def wrapper(*args, **kwargs):
            argnums = argnum if argnum is not None else list(range(len(args)))
            tensor_lists = tensor_list if tensor_list is not None else []

            if not isinstance(argnums, Sequence):
                argnums = [argnums]
            if not isinstance(tensor_lists, Sequence):
                tensor_lists = [tensor_lists]

            dispatch_args = []

            for a in argnums:
                # Only use extend if the marked argument really
                # is a (native) python Sequence
                if a in tensor_lists and isinstance(args[a], (list, tuple)):
                    dispatch_args.extend(args[a])
                else:
                    dispatch_args.append(args[a])

            interface = kwargs.pop("like", None)
            interface = interface or _multi_dispatch(dispatch_args)
            kwargs["like"] = interface

            return fn(*args, **kwargs)

        return wrapper

    return decorator


@multi_dispatch(argnum=[0], tensor_list=[0])
def block_diag(values, like=None):
    """Combine a sequence of 2D tensors to form a block diagonal tensor.

    Args:
        values (Sequence[tensor_like]): Sequence of 2D arrays/tensors to form
            the block diagonal tensor.

    Returns:
        tensor_like: the block diagonal tensor

    **Example**

    >>> t = [
    ...     np.array([[1, 2], [3, 4]]),
    ...     torch.tensor([[1, 2, 3], [-1, -6, -3]]),
    ...     torch.tensor(5)
    ... ]
    >>> qml.math.block_diag(t)
    tensor([[ 1,  2,  0,  0,  0,  0],
            [ 3,  4,  0,  0,  0,  0],
            [ 0,  0,  1,  2,  3,  0],
            [ 0,  0, -1, -6, -3,  0],
            [ 0,  0,  0,  0,  0,  5]])
    """
    values = np.coerce(values, like=like)
    return np.block_diag(values, like=like)


@multi_dispatch(argnum=[0], tensor_list=[0])
def concatenate(values, axis=0, like=None):
    """Concatenate a sequence of tensors along the specified axis.

    .. warning::

        Tensors that are incompatible (such as Torch and TensorFlow tensors)
        cannot both be present.

    Args:
        values (Sequence[tensor_like]): Sequence of tensor-like objects to
            concatenate. The objects must have the same shape, except in the dimension corresponding
            to axis (the first, by default).
        axis (int): The axis along which the input tensors are concatenated. If axis is None,
            tensors are flattened before use. Default is 0.

    Returns:
        tensor_like: The concatenated tensor.

    **Example**

    >>> x = tf.constant([0.6, 0.1, 0.6])
    >>> y = tf.Variable([0.1, 0.2, 0.3])
    >>> z = np.array([5., 8., 101.])
    >>> concatenate([x, y, z])
    <tf.Tensor: shape=(3, 3), dtype=float32, numpy=
    array([6.00e-01, 1.00e-01, 6.00e-01, 1.00e-01, 2.00e-01, 3.00e-01, 5.00e+00, 8.00e+00, 1.01e+02], dtype=float32)>
    """

    if like == "torch":
        import torch

        device = (
            "cuda"
            if any(
                t.device.type == "cuda" for t in values if isinstance(t, torch.Tensor)
            )
            else "cpu"
        )

        if axis is None:
            # flatten and then concatenate zero'th dimension
            # to reproduce numpy's behaviour
            values = [
                np.flatten(
                    torch.as_tensor(t, device=torch.device(device))
                )  # pragma: no cover
                for t in values
            ]
            axis = 0
        else:
            values = [
                torch.as_tensor(t, device=torch.device(device))
                for t in values  # pragma: no cover
            ]

    if like == "tensorflow" and axis is None:
        # flatten and then concatenate zero'th dimension
        # to reproduce numpy's behaviour
        values = [np.flatten(np.array(t)) for t in values]
        axis = 0

    return np.concatenate(values, axis=axis, like=like)


@multi_dispatch(argnum=[0], tensor_list=[0])
def diag(values, k=0, like=None):
    """Construct a diagonal tensor from a list of scalars.

    Args:
        values (tensor_like or Sequence[scalar]): sequence of numeric values that
            make up the diagonal
        k (int): The diagonal in question. ``k=0`` corresponds to the main diagonal.
            Use ``k>0`` for diagonals above the main diagonal, and ``k<0`` for
            diagonals below the main diagonal.

    Returns:
        tensor_like: the 2D diagonal tensor

    **Example**

    >>> x = [1., 2., tf.Variable(3.)]
    >>> qml.math.diag(x)
    <tf.Tensor: shape=(3, 3), dtype=float32, numpy=
    array([[1., 0., 0.],
           [0., 2., 0.],
           [0., 0., 3.]], dtype=float32)>
    >>> y = tf.Variable([0.65, 0.2, 0.1])
    >>> qml.math.diag(y, k=-1)
    <tf.Tensor: shape=(4, 4), dtype=float32, numpy=
    array([[0.  , 0.  , 0.  , 0.  ],
           [0.65, 0.  , 0.  , 0.  ],
           [0.  , 0.2 , 0.  , 0.  ],
           [0.  , 0.  , 0.1 , 0.  ]], dtype=float32)>
    >>> z = torch.tensor([0.1, 0.2])
    >>> qml.math.diag(z, k=1)
    tensor([[0.0000, 0.1000, 0.0000],
            [0.0000, 0.0000, 0.2000],
            [0.0000, 0.0000, 0.0000]])
    """
    if isinstance(values, (list, tuple)):
        values = np.stack(np.coerce(values, like=like), like=like)

    return np.diag(values, k=k, like=like)


@multi_dispatch(argnum=[0, 1])
def dot(tensor1, tensor2, like=None):
    """Returns the matrix or dot product of two tensors.

    * If both tensors are 0-dimensional, elementwise multiplication
      is performed and a 0-dimensional scalar returned.

    * If both tensors are 1-dimensional, the dot product is returned.

    * If the first array is 2-dimensional and the second array 1-dimensional,
      the matrix-vector product is returned.

    * If both tensors are 2-dimensional, the matrix product is returned.

    * Finally, if the the first array is N-dimensional and the second array
      M-dimensional, a sum product over the last dimension of the first array,
      and the second-to-last dimension of the second array is returned.

    Args:
        tensor1 (tensor_like): input tensor
        tensor2 (tensor_like): input tensor

    Returns:
        tensor_like: the matrix or dot product of two tensors
    """
    x, y = np.coerce([tensor1, tensor2], like=like)

    if like == "torch":
        if x.ndim == 0 and y.ndim == 0:
            return x * y

        if x.ndim <= 2 and y.ndim <= 2:
            return x @ y

        return np.tensordot(x, y, axes=[[-1], [-2]], like=like)

    if like == "tensorflow":
        if len(np.shape(x)) == 0 and len(np.shape(y)) == 0:
            return x * y

        if len(np.shape(y)) == 1:
            return np.tensordot(x, y, axes=[[-1], [0]], like=like)

        if len(np.shape(x)) == 2 and len(np.shape(y)) == 2:
            return x @ y

        return np.tensordot(x, y, axes=[[-1], [-2]], like=like)

    return np.dot(x, y, like=like)


@multi_dispatch(argnum=[0, 1])
def tensordot(tensor1, tensor2, axes=None, like=None):
    """Returns the tensor product of two tensors.
    In general ``axes`` specifies either the set of axes for both
    tensors that are contracted (with the first/second entry of ``axes``
    giving all axis indices for the first/second tensor) or --- if it is
    an integer --- the number of last/first axes of the first/second
    tensor to contract over.
    There are some non-obvious special cases:

    * If both tensors are 0-dimensional, ``axes`` must be 0.
      and a 0-dimensional scalar is returned containing the simple product.

    * If both tensors are 1-dimensional and ``axes=0``, the outer product
      is returned.

    * Products between a non-0-dimensional and a 0-dimensional tensor are not
      supported in all interfaces.

    Args:
        tensor1 (tensor_like): input tensor
        tensor2 (tensor_like): input tensor
        axes (int or list[list[int]]): Axes to contract over, see detail description.

    Returns:
        tensor_like: the tensor product of the two input tensors
    """
    tensor1, tensor2 = np.coerce([tensor1, tensor2], like=like)
    return np.tensordot(tensor1, tensor2, axes=axes, like=like)


@multi_dispatch(argnum=[0], tensor_list=[0])
def get_trainable_indices(values, like=None):
    """Returns a set containing the trainable indices of a sequence of
    values.

    Args:
        values (Iterable[tensor_like]): Sequence of tensor-like objects to inspect

    Returns:
        set[int]: Set containing the indices of the trainable tensor-like objects
        within the input sequence.

    **Example**

    >>> def cost_fn(params):
    ...     print("Trainable:", qml.math.get_trainable_indices(params))
    ...     return np.sum(np.sin(params[0] * params[1]))
    >>> values = [np.array([0.1, 0.2], requires_grad=True),
    ... np.array([0.5, 0.2], requires_grad=False)]
    >>> cost_fn(values)
    Trainable: {0}
    tensor(0.0899685, requires_grad=True)
    """
    trainable = requires_grad
    trainable_params = set()

    if like == "jax":
        import jax

        if not any(isinstance(v, jax.core.Tracer) for v in values):
            # No JAX tracing is occuring; treat all `DeviceArray` objects as trainable.

            # pylint: disable=function-redefined,unused-argument
            def trainable(p, **kwargs):
                return isinstance(p, jax.numpy.DeviceArray)

        else:
            # JAX tracing is occuring; use the default behaviour (only traced arrays
            # are treated as trainable). This is required to ensure that `jax.grad(func, argnums=...)
            # works correctly, as the argnums argnument determines which parameters are
            # traced arrays.
            trainable = requires_grad

    for idx, p in enumerate(values):
        if trainable(p, interface=like):
            trainable_params.add(idx)

    return trainable_params


def ones_like(tensor, dtype=None):
    """Returns a tensor of all ones with the same shape and dtype
    as the input tensor.

    Args:
        tensor (tensor_like): input tensor
        dtype (str, np.dtype, None): The desired output datatype of the array. If not provided, the dtype of
            ``tensor`` is used. This argument can be any supported NumPy dtype representation, including
            a string (``"float64"``), a ``np.dtype`` object (``np.dtype("float64")``), or
            a dtype class (``np.float64``). If ``tensor`` is not a NumPy array, the
            **equivalent** dtype in the dispatched framework is used.

    Returns:
        tensor_like: an all-ones tensor with the same shape and
        size as ``tensor``

    **Example**

    >>> x = torch.tensor([1., 2.])
    >>> ones_like(x)
    tensor([1, 1])
    >>> y = tf.Variable([[0], [5]])
    >>> ones_like(y, dtype=np.complex128)
    <tf.Tensor: shape=(2, 1), dtype=complex128, numpy=
    array([[1.+0.j],
           [1.+0.j]])>
    """
    if dtype is not None:
        return cast(np.ones_like(tensor), dtype)

    return np.ones_like(tensor)


@multi_dispatch(argnum=[0], tensor_list=[0])
def stack(values, axis=0, like=None):
    """Stack a sequence of tensors along the specified axis.

    .. warning::

        Tensors that are incompatible (such as Torch and TensorFlow tensors)
        cannot both be present.

    Args:
        values (Sequence[tensor_like]): Sequence of tensor-like objects to
            stack. Each object in the sequence must have the same size in the given axis.
        axis (int): The axis along which the input tensors are stacked. ``axis=0`` corresponds
            to vertical stacking.

    Returns:
        tensor_like: The stacked array. The stacked array will have one additional dimension
        compared to the unstacked tensors.

    **Example**

    >>> x = tf.constant([0.6, 0.1, 0.6])
    >>> y = tf.Variable([0.1, 0.2, 0.3])
    >>> z = np.array([5., 8., 101.])
    >>> stack([x, y, z])
    <tf.Tensor: shape=(3, 3), dtype=float32, numpy=
    array([[6.00e-01, 1.00e-01, 6.00e-01],
           [1.00e-01, 2.00e-01, 3.00e-01],
           [5.00e+00, 8.00e+00, 1.01e+02]], dtype=float32)>
    """
    values = np.coerce(values, like=like)
    return np.stack(values, axis=axis, like=like)


def einsum(indices, *operands, like=None):
    """Evaluates the Einstein summation convention on the operands.

    Args:
        indices (str): Specifies the subscripts for summation as comma separated list of
            subscript labels. An implicit (classical Einstein summation) calculation is
            performed unless the explicit indicator ‘->’ is included as well as subscript
            labels of the precise output form.
        operands (tuple[tensor_like]): The tensors for the operation.

    Returns:
        tensor_like: The calculation based on the Einstein summation convention.

    **Examples**

    >>> a = np.arange(25).reshape(5,5)
    >>> b = np.arange(5)
    >>> c = np.arange(6).reshape(2,3)

    Trace of a matrix:

    >>> qml.math.einsum('ii', a)
    60

    Extract the diagonal (requires explicit form):

    >>> qml.math.einsum('ii->i', a)
    array([ 0,  6, 12, 18, 24])

    Sum over an axis (requires explicit form):

    >>> qml.math.einsum('ij->i', a)
    array([ 10,  35,  60,  85, 110])

    Compute a matrix transpose, or reorder any number of axes:

    >>> np.einsum('ij->ji', c)
    array([[0, 3],
           [1, 4],
           [2, 5]])

    Matrix vector multiplication:

    >>> np.einsum('ij,j', a, b)
    array([ 30,  80, 130, 180, 230])
    """
    if like is None:
        like = _multi_dispatch(operands)
    operands = np.coerce(operands, like=like)
    return np.einsum(indices, *operands, like=like)


def where(condition, x=None, y=None):
    """Returns elements chosen from x or y depending on a boolean tensor condition,
    or the indices of entries satisfying the condition.

    The input tensors ``condition``, ``x``, and ``y`` must all be broadcastable to the same shape.

    Args:
        condition (tensor_like[bool]): A boolean tensor. Where ``True`` , elements from
            ``x`` will be chosen, otherwise ``y``. If ``x`` and ``y`` are ``None`` the
            indices where ``condition==True`` holds will be returned.
        x (tensor_like): values from which to choose if the condition evaluates to ``True``
        y (tensor_like): values from which to choose if the condition evaluates to ``False``

    Returns:
        tensor_like or tuple[tensor_like]: If ``x is None`` and ``y is None``, a tensor
        or tuple of tensors with the indices where ``condition`` is ``True`` .
        Else, a tensor with elements from ``x`` where the ``condition`` is ``True``,
        and ``y`` otherwise. In this case, the output tensor has the same shape as
        the input tensors.

    **Example with three arguments**

    >>> a = torch.tensor([0.6, 0.23, 0.7, 1.5, 1.7], requires_grad=True)
    >>> b = torch.tensor([-1., -2., -3., -4., -5.], requires_grad=True)
    >>> math.where(a < 1, a, b)
    tensor([ 0.6000,  0.2300,  0.7000, -4.0000, -5.0000], grad_fn=<SWhereBackward>)

    .. warning::

        The output format for ``x=None`` and ``y=None`` follows the respective
        interface and differs between TensorFlow and all other interfaces:
        For TensorFlow, the output is a tensor with shape
        ``(num_true, len(condition.shape))`` where ``num_true`` is the number
        of entries in ``condition`` that are ``True`` .
        The entry at position ``(i, j)`` is the ``j`` th entry of the ``i`` th
        index.
        For all other interfaces, the output is a tuple of tensor-like objects,
        with the ``j`` th object indicating the ``j`` th entries of all indices.
        Also see the examples below.

    **Example with single argument**

    For Torch, Autograd, JAX and NumPy, the output formatting is as follows:

    >>> a = [[0.6, 0.23, 1.7],[1.5, 0.7, -0.2]]
    >>> math.where(torch.tensor(a) < 1)
    (tensor([0, 0, 1, 1]), tensor([0, 1, 1, 2]))

    This is not a single tensor-like object but corresponds to the shape
    ``(2, 4)`` . For TensorFlow, on the other hand:

    >>> math.where(tf.constant(a) < 1)
    tf.Tensor(
    [[0 0]
     [0 1]
     [1 1]
     [1 2]], shape=(4, 2), dtype=int64)

    As we can see, the dimensions are swapped and the output is a single Tensor.
    Note that the number of dimensions of the output does *not* depend on the input
    shape, it is always two-dimensional.

    """
    if x is None and y is None:
        interface = _multi_dispatch([condition])
        res = np.where(condition, like=interface)

        if interface == "tensorflow":
            return np.transpose(np.stack(res))

        return res

    interface = _multi_dispatch([condition, x, y])
    res = np.where(condition, x, y, like=interface)

    return res


@multi_dispatch(argnum=[0, 1])
def frobenius_inner_product(A, B, normalize=False, like=None):
    r"""Frobenius inner product between two matrices.

    .. math::

        \langle A, B \rangle_F = \sum_{i,j=1}^n A_{ij} B_{ij} = \operatorname{tr} (A^T B)

    The Frobenius inner product is equivalent to the Hilbert-Schmidt inner product for
    matrices with real-valued entries.

    Args:
        A (tensor_like[float]): First matrix, assumed to be a square array.
        B (tensor_like[float]): Second matrix, assumed to be a square array.
        normalize (bool): If True, divide the inner product by the Frobenius norms of A and B.

    Returns:
        float: Frobenius inner product of A and B

    **Example**

    >>> A = np.random.random((3,3))
    >>> B = np.random.random((3,3))
    >>> qml.math.frobenius_inner_product(A, B)
    3.091948202943376
    """
    A, B = np.coerce([A, B], like=like)

    inner_product = np.sum(A * B)

    if normalize:
        norm = np.sqrt(np.sum(A * A) * np.sum(B * B))
        inner_product = inner_product / norm

    return inner_product


@multi_dispatch(argnum=[1])
def scatter(indices, array, new_dims, like=None):
    """Scatters an array into a tensor of shape new_dims according to indices.

    This operation is similar to scatter_element_add, except that the tensor
    is zero-initialized. Calling scatter(indices, array, new_dims) is identical
    to calling scatter_element_add(np.zeros(new_dims), indices, array)

    Args:
        indices (tensor_like[int]): Indices to update
        array (tensor_like[float]): Values to assign to the new tensor
        new_dims (int or tuple[int]): The shape of the new tensor
        like (str): Manually chosen interface to dispatch to.
    Returns:
        tensor_like[float]: The tensor with the values modified the given indices.

    **Example**

    >>> indices = np.array([4, 3, 1, 7])
    >>> updates = np.array([9, 10, 11, 12])
    >>> shape = 8
    >>> qml.math.scatter(indices, updates, shape)
    array([ 0, 11,  0, 10,  9,  0,  0, 12])
    """
    return np.scatter(indices, array, new_dims, like=like)


@multi_dispatch(argnum=[0, 2])
def scatter_element_add(tensor, index, value, like=None):
    """In-place addition of a multidimensional value over various
    indices of a tensor.

    Args:
        tensor (tensor_like[float]): Tensor to add the value to
        index (tuple or list[tuple]): Indices to which to add the value
        value (float or tensor_like[float]): Value to add to ``tensor``
        like (str): Manually chosen interface to dispatch to.
    Returns:
        tensor_like[float]: The tensor with the value added at the given indices.

    **Example**

    >>> tensor = torch.tensor([[0.1, 0.2, 0.3], [0.4, 0.5, 0.6]])
    >>> index = (1, 2)
    >>> value = -3.1
    >>> qml.math.scatter_element_add(tensor, index, value)
    tensor([[ 0.1000,  0.2000,  0.3000],
            [ 0.4000,  0.5000, -2.5000]])

    If multiple indices are given, in the form of a list of tuples, the
    ``k`` th tuple is interpreted to contain the ``k`` th entry of all indices:

    >>> indices = [(1, 0), (2, 1)] # This will modify the entries (1, 2) and (0, 1)
    >>> values = torch.tensor([10, 20])
    >>> qml.math.scatter_element_add(tensor, indices, values)
    tensor([[ 0.1000, 20.2000,  0.3000],
            [ 0.4000,  0.5000, 10.6000]])
    """
    if len(np.shape(tensor)) == 0 and index == ():
        return tensor + value

    return np.scatter_element_add(tensor, index, value, like=like)


def unwrap(values, max_depth=None):
    """Unwrap a sequence of objects to NumPy arrays.

    Note that tensors on GPUs will automatically be copied
    to the CPU.

    Args:
        values (Sequence[tensor_like]): sequence of tensor-like objects to unwrap
        max_depth (int): Positive integer indicating the depth of unwrapping to perform
            for nested tensor-objects. This argument only applies when unwrapping
            Autograd ``ArrayBox`` objects.

    **Example**

    >>> values = [np.array([0.1, 0.2]), torch.tensor(0.1, dtype=torch.float64), torch.tensor([0.5, 0.2])]
    >>> math.unwrap(values)
    [array([0.1, 0.2]), 0.1, array([0.5, 0.2], dtype=float32)]

    This function will continue to work during backpropagation:

    >>> def cost_fn(params):
    ...     unwrapped_params = math.unwrap(params)
    ...     print("Unwrapped:", [(i, type(i)) for i in unwrapped_params])
    ...     return np.sum(np.sin(params))
    >>> params = np.array([0.1, 0.2, 0.3])
    >>> grad = autograd.grad(cost_fn)(params)
    Unwrapped: [(0.1, <class 'float'>), (0.2, <class 'float'>), (0.3, <class 'float'>)]
    >>> print(grad)
    [0.99500417 0.98006658 0.95533649]
    """
    res = []

    for t in values:
        if isinstance(t, ArrayBox):
            a = np.to_numpy(t, max_depth=max_depth)
        else:
            a = np.to_numpy(t)

        if isinstance(a, ndarray) and not a.shape:
            # if NumPy array is scalar, convert to a Python float
            res.append(a.tolist())
        else:
            res.append(a)

    return res


def add(*args, **kwargs):
    """Add arguments element-wise."""
    try:
        return np.add(*args, **kwargs)
    except TypeError:
        # catch arg1 = torch, arg2=numpy error
        # works fine with opposite order
        return np.add(args[1], args[0], *args[2:], **kwargs)


@multi_dispatch()
def iscomplex(tensor, like=None):
    """Return True if the tensor has a non-zero complex component."""
    if like == "tensorflow":
        import tensorflow as tf

        imag_tensor = tf.math.imag(tensor)
        return tf.math.count_nonzero(imag_tensor) > 0

    if like == "torch":
        import torch

        if torch.is_complex(tensor):
            imag_tensor = torch.imag(tensor)
            return torch.count_nonzero(imag_tensor) > 0
        return False

    return np.iscomplex(tensor)


@multi_dispatch()
def expm(tensor, like=None):
    """Compute the matrix exponential of an array :math:`e^{X}`.

    ..note::
        This function is not differentiable with Autograd, as it
        relies on the scipy implementation.
    """
    if like == "torch":
        return tensor.matrix_exp()
    if like == "jax":
        from jax.scipy.linalg import expm as jax_expm

        return jax_expm(tensor)
    if like == "tensorflow":
        import tensorflow as tf

        return tf.linalg.expm(tensor)
    from scipy.linalg import expm as scipy_expm

    return scipy_expm(tensor)



import numpy as np
from scipy.optimize import OptimizeResult


def newton_descent(
    fun,
    x0,
    args=(),
    maxfev=None,
    stepsize=0.01,
    maxiter=100,
    tol=10 ** (-6),
    jac=None,
    hess=None,
    callback=None,
    **options
):
    """
    Minimize a function `fun` using Newton's method, a second order method.
    scipy.optimize.minimize compatible implementation of Newton's method for `method` == 'newton'.

    PARAMETERS
    ----------
    fun : callable
        Function to minimize
    x0 : ndarray
        Initial guess.
    args : sequence, optional
        Arguments to pass to `func`.
    maxfev : int, optional
        Maximum number of function evaluations.
    stepsize : float
        Step size of each gradient descent step.
    maxiter : int, optional
        Maximum number of iterations.
    tol : float
        Tolerance before the optimizer terminates; if `tol` is larger
        than the difference between two steps, terminate optimization.
    jac : callable
        Callable gradient function.
    hess : callable
        Callable hessian function.
    callback : callable, optional
        Called after each iteration, as ``callback(xk)``, where ``xk`` is the
        current parameter vector.

    RETURNS
    -------
    OptimizeResult : OptimizeResult
        Scipy OptimizeResult object.
    """

    assert jac is not None, "This method needs jacobian specified"
    assert hess is not None, "This method needs hessian specified"

    bestx = x0
    besty = fun(x0)
    funcalls = 1
    niter = 0
    improved = True
    stop = False

    testx = np.copy(bestx)
    testy = np.real(fun(testx, *args))
    while improved and not stop and niter < maxiter:
        improved = False
        niter += 1

        # Scale gradient with inverse of Hessian
        scaled_gradient = np.linalg.solve(hess(testx), jac(testx))

        # compute Newton descent step
        testx = testx - stepsize * scaled_gradient
        testy = np.real(fun(testx, *args))

        if np.abs(besty - testy) < tol:
            improved = False
        else:
            # if testy < besty:
            besty = testy
            bestx = testx
            improved = True

        if callback is not None:
            callback(bestx)
        if maxfev is not None and funcalls >= maxfev:
            stop = True
            break

    return OptimizeResult(
        fun=besty, x=bestx, nit=niter, nfev=funcalls, success=(niter > 1)
    )



import numpy as np
from typing import List
from boto3.session import Session
from botocore.exceptions import NoRegionError
from braket.aws import AwsDevice
from braket.aws.aws_session import AwsSession

from openqaoa.backends.devices_core import DeviceBase


class DeviceAWS(DeviceBase):

    """
    Contains the required information and methods needed to access QPUs hosted
    on AWS Braket.

    Attributes
    ----------
    available_qpus: `list`
        When connection to AWS is established, this attribute contains a list
        of device names which can be used to access the selected device by
        reinitialising the Access Object with the name of the available device
        as input to the device_name parameter.
    n_qubits: `int`
        The maximum number of qubits available for the selected backend. Only
        available if check_connection method is executed and a connection to the
        qpu and provider is established.
    """

    def __init__(
        self,
        device_name: str,
        s3_bucket_name: str = None,
        aws_region: str = None,
        folder_name: str = "openqaoa",
    ):
        """Input the device arn and the name of the folder in which all the
        results for the QPU runs would be saved on the pre-defined s3 bucket.
        Note that the user is required to authenticate through the AWS CLI
        before being able to use this Device object.

        See: https://docs.aws.amazon.com/cli/latest/userguide/cli-configure-files.html for further details.

        Parameters
        ----------
        device_name: `str`
                The ARN string of the braket QPU/simulator to be used
        s3_bucket_name: `str`
            The name of S3 Bucket where the Braket run results will be saved.
        aws_region: `str`
            The aws region in which the QPU/simulator is located. Defaults to the
            region set in aws cli config.
        folder_name: `str`
            The name of the folder in the s3 bucket that will contain the results
            from the tasks performed in this run.
        """

        self.device_name = device_name
        self.device_location = "aws"
        self.s3_bucket_name = s3_bucket_name
        self.aws_region = aws_region
        self.folder_name = folder_name

        self.provider_connected = None
        self.qpu_connected = None

    def check_connection(self) -> bool:
        self.provider_connected = self._check_provider_connection()

        if self.provider_connected == False:
            return self.provider_connected

        # Only QPUs that are available for the specified aws region on Braket
        # will be shown. We filter out QPUs that do not work with the circuit model
        sess_devices = self.aws_session.search_devices()

        device_filter = np.multiply(
            [each_dict["deviceStatus"] == "ONLINE" for each_dict in sess_devices],
            [
                each_dict["providerName"] != "D-Wave Systems"
                for each_dict in sess_devices
            ],
        )
        active_devices = np.array(sess_devices)[device_filter].tolist()

        self.available_qpus = [
            backend_dict["deviceArn"] for backend_dict in active_devices
        ]

        if self.device_name == "":
            return self.provider_connected

        self.qpu_connected = self._check_backend_connection()

        if self.provider_connected and self.qpu_connected:
            return True
        else:
            return False

    def _check_backend_connection(self) -> bool:
        if self.device_name in self.available_qpus:
            self.backend_device = AwsDevice(self.device_name, self.aws_session)
        else:
            print(
                """
                These are the only available devices for this aws region: 
                {}. Try a different aws region if the device you are looking 
                for is not in the list.'
                """.format(
                    self.available_qpus
                )
            )
            return False

        # Get the maximum number of qubits for that particular AWS Backend
        try:
            self.n_qubits = self.backend_device.properties.paradigm.qubitCount
        except AttributeError:
            print(
                "OpenQAOA is unable to retrieve the number of qubits available in the selected QPU."
            )
            return False
        else:
            return True

    def _check_provider_connection(self) -> bool:
        try:
            sess = Session(region_name=self.aws_region)
            self.aws_session = AwsSession(sess, default_bucket=self.s3_bucket_name)
            self.aws_region = self.aws_session.region
            self.s3_bucket_name = self.aws_session.default_bucket()
            return True
        except NoRegionError:
            self.aws_session = None
            return True
        except Exception as e:
            print(
                "An Exception has occured when trying to connect with the provider."
                "You are required to authenticate through the AWS CLI in order to connect to the Braket QPUs."
                "Please check if you have properly set it up."
                "See: https://docs.aws.amazon.com/cli/latest/userguide/cli-configure-files.html for further details. : {}".format(
                    e
                )
            )
            return False

    def connectivity(self) -> List[List[int]]:
        return self.backend_device.topology_graph.edges
